from typing import Dict, List, Optional
from jina.executors.segmenters import BaseSegmenter
from jina.executors.decorators import single


class NLTKSentencizer(BaseSegmenter):
    """
    Segment text into sentences using :class:`nltk.PunktSentenceTokenizer`
    for the specified language.

    Example:

    >>> sentencizer = NLTKSentencizer() # doctest:+ELLIPSIS
    NLTKSentencizer@4...[W]:__main__.NLTKSentencizer works on keys, but no keys are specified
    NLTKSentencizer@...[I]:post_init may take some time...
    NLTKSentencizer@...[I]:NLTK english sentence tokenizer ready.
    NLTKSentencizer@...[I]:post_init may take some time takes ... seconds (...s)
    >>> text = "Today is a good day. Can't wait for tomorrow!"
    >>> sentencizer.segment([text])
    [{'text': 'Today is a good day.', 'offset': 0, 'location': [0, 20]}, {'text': "Can't wait for tomorrow!", 'offset': 1, 'location': [21, 45]}]


    :param language: default='english'. Lowercased language name to initialize the sentence tokenizer, accepted languages are listed in :attr:`SUPPORTED_LANGUAGES`.
    :type language: Optional[str]
    :param resource_url: default=None. Optional parameter to specify the location for tokenizer model, overrides `language`. For a file path, should follow `file://path_to_resource` format. For more information, see https://www.nltk.org/data.html
    :type resource_url: Optional[str]
    :param args: Additional positional arguments
    :param kwargs: Additional keyword arguments
    """

    SUPPORTED_LANGUAGES = [
        'czech',
        'danish',
        'dutch',
        'english',
        'estonian',
        'finnish',
        'french',
        'german',
        'greek',
        'italian',
        'norwegian',
        'polish',
        'portuguese',
        'russian',
        'slovene',
        'spanish',
        'swedish',
        'turkish',
    ]

    def __init__(
        self,
        language: Optional[str] = 'english',
        resource_url: Optional[str] = None,
        *args,
        **kwargs,
    ):
        """Set constructor"""
        super().__init__(*args, **kwargs)
        if resource_url:
            self.resource_url = resource_url
            self.language = None
        else:
            self.resource_url = f'tokenizers/punkt/{language}.pickle'
            self.language = language

    def post_init(self):
        from nltk.data import load

        try:
            self._sent_tokenizer = load(self.resource_url)
        except LookupError:
            if self.language in self.SUPPORTED_LANGUAGES:
                try:
                    import nltk

                    nltk.download('punkt')
                    self._sent_tokenizer = load(self.resource_url)
                except:
                    self.logger.error(
                        f'Please ensure that "nltk_data" folder is accessible to your working directory.'
                    )
                    raise
            elif self.language:
                self.logger.error(
                    f'The language you specified ("{self.language}") is not supplied by NLTK.'
                    f'Please ensure that language is one of the acceptable languages listed in '
                    f'{self.SUPPORTED_LANGUAGES}.\nFor latest list of supported languages check out '
                    f'https://raw.githubusercontent.com/nltk/nltk_data/gh-pages/index.xml with '
                    f'id="punkt"'
                )
                raise
            else:
                self.logger.error(
                    f'The resource_url you provided could not be loaded: {self.resource_url}'
                )
                raise
        self.logger.info(f'NLTK {self.language} sentence tokenizer ready.')

    @single
    def segment(self, text: str, *args, **kwargs) -> List[Dict]:
        """
        Segment text into sentences.

        :param text: The text to be sentencized.
        :type text: str
        :param args: Additional positional arguments
        :param kwargs: Additional keyword arguments
        :return: List of dictonaries representing sentences with three keys: ``text``, for representing the text of the sentence; ``offset``, representing the order of the sentence within input text; ``location``, a list with start and end indeces for the sentence.
        :rtype: List[Dict]
        """
        sentences = self._sent_tokenizer.tokenize(text)
        results = []
        start = 0
        for i, s in enumerate(sentences):
            start = text[start:].find(s) + start
            end = start + len(s)
            results.append({'text': s, 'offset': i, 'location': [start, end]})
            start = end

        return results



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import numpy as np
import torch

from jina.executors.decorators import batching, as_ndarray
from jina.executors.encoders.frameworks import BaseTorchEncoder
from jina.executors.devices import TorchDevice


class FaceNetEncoder(BaseTorchEncoder, TorchDevice):
    """FaceNetEncoder encodes face images into an embedding with 512 dimensions.
    Works best with outputs from FaceNetSegmenter (e.g. images with faces, 160x160, normalized).

    - Input shape: `BatchSize x (Channels x Height x Width)`
    - Output shape: `BatchSize x 512`

    `Channels` dimension can be changed (e.g. set `channel_axis` to 1 for channels-first mode instead of channels-last).

    :param pretrained_weights: Weights to use for face embedder. Options: "vggface2", "casia-webface"
    :param channel_axis: Axis of channels in the image. Default is 3 (channels-last), use 1 for channels-first.
    """

    def __init__(self,
                 pretrained_weights: str = 'vggface2',
                 channel_axis: int = 1,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pretrained_weights = pretrained_weights
        self.channel_axis = channel_axis

        self._default_channel_axis = 1

    def post_init(self):
        from facenet_pytorch import InceptionResnetV1

        self.model = InceptionResnetV1(pretrained=self.pretrained_weights,
                                       classify=False,
                                       device=self.device).eval()
        self.model.to(self.device)

    @batching
    @as_ndarray
    def encode(self, content: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        """Transform a numpy `ndarray` of shape `BatchSize x (Height x Width x Channel)`
        into a numpy `ndarray` of shape `Batchsize x EmbeddingDimension`.

        :param content: A numpy `ndarray` of strings.
        :param args: Additional positional arguments.
        :param kwargs: Additional positional arguments.
        :return: A `BatchSize x EmbeddingDimension` numpy array.
        """
        if self.channel_axis != self._default_channel_axis:
            content = np.moveaxis(content, self.channel_axis, self._default_channel_axis)

        with torch.no_grad():
            images = torch.from_numpy(content.astype('float32')).to(self.device)
            embedded_faces = self.model(images)
        return embedded_faces.detach().cpu()



from collections import deque
from itertools import islice
from typing import Dict, List

from jina.executors.decorators import single
from jina.executors.segmenters import BaseSegmenter


class SlidingWindowSegmenter(BaseSegmenter):
    """
    :class:`SlidingWindowSegmenter` split the text on the doc-level
    into overlapping substrings on the chunk-level.
    The text is split into substrings of length ``window_size`` if possible.
    The degree of overlapping can be configured through the ``step_size`` parameter.
    The substrings that are shorter than the ``min_substring_len`` will be discarded.

    :param window_size: the window size that will be used to split the text into substrings
    :param step_size: the size of the degree of overlapping
    :param min_substring_len: the minimum size of substrings. Anything shorter will be discarded
    """

    def __init__(self,
                 window_size: int = 300,
                 step_size: int = 150,
                 min_substring_len: int = 1,
                 *args, **kwargs):
        """Set constructor."""
        super().__init__(*args, **kwargs)
        self.window_size = window_size
        self.step_size = step_size
        self.min_substring_len = min_substring_len
        if self.min_substring_len > self.window_size:
            self.logger.warning(
                'the min_substring_len (={}) should be smaller to the window_size (={})'.format(
                    self.min_substring_len, self.window_size))
        if self.window_size <= 0:
            self.logger.warning(
                f'the window_size (={self.window_size}) should be larger than zero')
        if self.step_size > self.window_size:
            self.logger.warning(
                'the step_size (={}) should not be larger than the window_size (={})'.format(
                    self.window_size, self.step_size))

    @single
    def segment(self, text: str, *args, **kwargs) -> List[Dict]:
        """
        Split the text into overlapping chunks
        :param text: the raw text in string format
        :return: a list of chunk dicts
        """

        def sliding_window(size, step):
            i = iter(text)
            d = deque(islice(i, size),
                      maxlen=size)
            if not d:
                # empty text
                return results
            while True:
                yield iter(d)
                try:
                    d.append(next(i))
                except StopIteration:
                    return
                d.extend(next(i, None)
                         for _ in range(step - 1))

        chunks = [''.join(filter(None, list(chunk))) for chunk in
                  sliding_window(self.window_size, self.step_size)]
        results = []
        for idx, s in enumerate(chunks):
            if self.min_substring_len <= len(s):
                results.append(dict(
                    text=s,
                    offset=idx,
                    weight=1.0))
        return results



import os
from typing import Dict, Optional, Tuple, List, Union

import numpy as np

from jina.executors.decorators import batching
from jina.executors.rankers import Match2DocRanker
from jina.excepts import PretrainedModelFileDoesNotExist

if False:
    import lightgbm


class LightGBMRanker(Match2DocRanker):
    """
    Computes a relevance score for each match using a pretrained Ltr model trained with LightGBM (https://lightgbm.readthedocs.io/en/latest/index.html)

    :param model_path: path to the pretrained model previously trained using LightGBM
    :param query_feature_names: name of the features to extract from query Documents and used to compute relevance scores by the model loaded
    from model_path
    :param match_feature_names: name of the features to extract from match Documents and used to compute relevance scores by the model loaded
    from model_path
    :param query_categorical_features: name of features contained in `query_feature_names` corresponding to categorical features.
    :param match_categorical_features: name of features contained in `match_feature_names` corresponding to categorical features.
    :param query_features_before: True if `query_feature_names` must be placed before the `match` ones in the `dataset` used for prediction.
    :param args: Additional positional arguments
    :param kwargs: Additional keyword arguments

    .. note::
        The name of the features are used to extract the features from incoming `documents`. Check how these features are accessed in
        :class:`Document` at https://docs.jina.ai/api/jina.types.document/

    """

    def __init__(self,
                 model_path: Optional[str] = 'tmp/model.txt',
                 query_feature_names: Tuple[str] = ['tags__query_length', 'tags__query_language'],
                 match_feature_names: Tuple[str] = ['tags__document_length', 'tags__document_language',
                                                    'tags__document_pagerank'],
                 query_categorical_features: Optional[List[str]] = None,
                 match_categorical_features: Optional[List[str]] = None,
                 query_features_before: bool = True,
                 *args,
                 **kwargs):
        super().__init__(query_required_keys=query_feature_names, match_required_keys=match_feature_names, *args,
                         **kwargs)
        self.model_path = model_path
        self.query_feature_names = query_feature_names
        self.match_feature_names = match_feature_names
        self.query_categorical_features = query_categorical_features
        self.match_categorical_features = match_categorical_features
        self.query_features_before = query_features_before

    def post_init(self):
        super().post_init()
        if self.model_path and os.path.exists(self.model_path):
            import lightgbm
            self.booster = lightgbm.Booster(model_file=self.model_path)
            model_num_features = self.booster.num_feature()
            expected_num_features = len(self.query_feature_names + self.match_feature_names)
            if model_num_features != expected_num_features:
                raise ValueError(
                    f'The number of features expected by the LightGBM model {model_num_features} is different'
                    f'than the ones provided in input {expected_num_features}')
        else:
            raise PretrainedModelFileDoesNotExist(f'model {self.model_path} does not exist')

    def _get_features_dataset(self, query_meta: List[Dict], match_meta: List[List[Dict]]) -> 'lightgbm.Dataset':
        import lightgbm

        def _get_features_per_query(q_meta, m_meta):
            query_features = np.array(
                [[q_meta[feat] for feat in self.query_feature_names] for _ in range(0, len(m_meta))])
            match_features = np.array([[meta[feat] for feat in self.match_feature_names] for meta in m_meta])
            return query_features, match_features

        q_features, m_features = [], []
        for q_meta, m_meta in zip(query_meta, match_meta):
            q_f, m_f = _get_features_per_query(q_meta, m_meta)
            q_features.append(q_f)
            m_features.append(m_f)

        query_features = np.vstack(q_features)
        query_dataset = lightgbm.Dataset(data=query_features, feature_name=self.query_feature_names,
                                         categorical_feature=self.query_categorical_features, free_raw_data=False)

        match_features = np.vstack(m_features)
        match_dataset = lightgbm.Dataset(data=match_features, feature_name=self.match_feature_names,
                                         categorical_feature=self.match_categorical_features, free_raw_data=False)
        if self.query_features_before:
            return query_dataset.construct().add_features_from(match_dataset.construct())
        else:
            return match_dataset.construct().add_features_from(query_dataset.construct())

    @batching(slice_nargs=3)
    def score(
            self, old_match_scores: List[List[float]], query_meta: List[Dict], match_meta: List[List[Dict]]
    ) -> 'np.ndarray':
        """
            Computes a relevance score for each match using a pretrained Ltr model trained with LightGBM (https://lightgbm.readthedocs.io/en/latest/index.html)

        :param query_meta: List of features extracted from query `Document`. Extracted according to `query_feature_names`.
            The length of the list is the number of features for which the scores will be computed
        :param old_match_scores: Previous scores assigned to query match pairs, unused in this case.
        :param match_meta: List of list of features extracted from match `Document`. Extracted according to `match_feature_names`.
            The list's length is equal to the number of queries, and each element is a list of the number of matches.
        :return:
        """

        dataset = self._get_features_dataset(query_meta=query_meta, match_meta=match_meta)
        return self.booster.predict(dataset.get_data())



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import argparse
from argparse import Namespace
from typing import Optional, Dict, List, Union, Set

from .. import Pea
from .. import Pod
from .. import BasePod
from ...enums import PollingType, PeaRoleType


class CompoundPod(BasePod):
    """A CompoundPod is a immutable set of pods, which run in parallel. They share the same input and output socket.
    Internally, the peas of the pods can run with the process/thread backend. They can be also run in their own containers.
    :param args: pod arguments parsed from the CLI
    :param needs: pod names of preceding pods, the output of these pods are going into the input of this pod
    """

    def __init__(self, args: Union['argparse.Namespace', Dict], needs: Set[str] = None):
        super().__init__(args, needs)
        self.replica_list = []  # type: List['Pod']
        if isinstance(args, Dict):
            # This is used when a Pod is created in a remote context, where peas & their connections are already given.
            self.replicas_args = args
        else:
            self.replicas_args = self._parse_args(args)

    @property
    def port_expose(self) -> int:
        """Get the grpc port number


        .. # noqa: DAR201
        """
        return self.head_args.port_expose

    @property
    def host(self) -> str:
        """Get the host name of this Pod


        .. # noqa: DAR201
        """
        return self.head_args.host

    def _parse_args(
        self, args: Namespace
    ) -> Dict[str, Optional[Union[List[Namespace], Namespace]]]:
        return self._parse_base_pod_args(
            args,
            attribute='replicas',
            id_attribute_name='replica_id',
            role_type=PeaRoleType.REPLICA,
            repetition_attribute='replicas',
            polling_type=PollingType.ANY,
        )

    @property
    def head_args(self):
        """
        Get the arguments for the `head` of this BasePod.

        :return: arguments of the head pea
        """
        return self.replicas_args['head']

    @head_args.setter
    def head_args(self, args):
        """
        Set the arguments for the `head` of this BasePod.

        :param args: arguments of the head pea
        """
        self.replicas_args['head'] = args

    @property
    def tail_args(self):
        """
        Get the arguments for the `tail` of this BasePod.

        :return: arguments of the tail pea
        """
        return self.replicas_args['tail']

    @tail_args.setter
    def tail_args(self, args):
        """
        Set the arguments for the `tail` of this BasePod.

        :param args: arguments of the tail pea
        """
        self.replicas_args['tail'] = args

    @property
    def all_args(
        self,
    ) -> Dict[
        str,
        Union[
            List[Union[List[Namespace], Namespace, None]],
            list,
            List[Namespace],
            Namespace,
            None,
        ],
    ]:
        """
        Get all arguments of all Peas and Pods (replicas) in this CompoundPod.

        :return: arguments for all Peas and pods
        """
        args = {
            'peas': ([self.replicas_args['head']] if self.replicas_args['head'] else [])
            + ([self.replicas_args['tail']] if self.replicas_args['tail'] else []),
            'replicas': self.replicas_args['replicas'],
        }
        return args

    @property
    def num_peas(self) -> int:
        """
        Get the number of running :class:`Pod`

        :return: total number of peas including head and tail
        """
        return sum([replica.num_peas for replica in self.replica_list]) + len(self.peas)

    def __eq__(self, other: 'CompoundPod'):
        return self.num_peas == other.num_peas and self.name == other.name

    def start(self) -> 'CompoundPod':
        """
        Start to run all :class:`Pod` and :class:`Pea` in this CompoundPod.

        :return: started CompoundPod

        .. note::
            If one of the :class:`Pod` fails to start, make sure that all of them
            are properly closed.
        """
        if getattr(self.args, 'noblock_on_start', False):
            for _args in self.all_args['peas']:
                _args.noblock_on_start = True
                self._enter_pea(Pea(_args))
            for _args in self.all_args['replicas']:
                _args.noblock_on_start = True
                _args.polling = PollingType.ALL
                self._enter_replica(Pod(_args))

            # now rely on higher level to call `wait_start_success`
            return self
        else:
            try:
                for _args in self.all_args['peas']:
                    self._enter_pea(Pea(_args))
                for _args in self.all_args['replicas']:
                    self._enter_replica(Pod(_args))
            except:
                self.close()
                raise
            return self

    def wait_start_success(self) -> None:
        """
        Block until all pods and peas start successfully.
        If not successful, it will raise an error hoping the outer function to catch it
        """

        if not self.args.noblock_on_start:
            raise ValueError(
                f'{self.wait_start_success!r} should only be called when `noblock_on_start` is set to True'
            )

        try:
            for p in self.peas:
                p.wait_start_success()
            for p in self.replica_list:
                p.wait_start_success()
        except:
            self.close()
            raise

    def _enter_replica(self, replica: 'Pod') -> None:
        self.replica_list.append(replica)
        self.enter_context(replica)

    def join(self):
        """Wait until all pods and peas exit."""
        try:
            for p in self.peas:
                p.join()
            for p in self.replica_list:
                p.join()
        except KeyboardInterrupt:
            pass
        finally:
            self.peas.clear()
            self.replica_list.clear()

    @property
    def is_ready(self) -> bool:
        """
        Checks if Pod is read.
        :return: true if the peas and pods are ready to serve requests

        .. note::
            A Pod is ready when all the Peas it contains are ready
        """
        return all(
            [p.is_ready.is_set() for p in self.peas]
            + [p.is_ready for p in self.replica_list]
        )

    def _set_after_to_pass(self, args):
        if PollingType.ANY.is_push:
            # ONLY reset when it is push
            args.uses_after = '_pass'

    def rolling_update(self):
        """
        Update all pods of this compound pod.
        """
        for i in range(len(self.replica_list)):
            replica = self.replica_list[i]
            replica.close()
            _args = self.all_args['replicas'][i]
            _args.noblock_on_start = False
            new_replica = Pod(_args)
            self.enter_context(new_replica)
            self.replica_list[i] = new_replica



import inspect
from typing import Dict, List, Union

from .. import BaseExecutor
from ...helper import typename


class BaseSegmenter(BaseExecutor):
    """:class:`BaseSegmenter` works on doc-level,
    it chunks Documents into set of Chunks
    :param args: Variable length arguments
    :param kwargs: Variable length keyword arguments
    """

    def segment(self, *args, **kwargs) -> Union[List[List[Dict]], List[Dict]]:
        """
        :param args: Variable length arguments
        :param kwargs: Variable length keyword arguments
        """
        raise NotImplementedError



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import os
import numpy as np

from jina.executors.encoders import BaseEncoder
from jina.excepts import PretrainedModelFileDoesNotExist

cur_dir = os.path.dirname(os.path.abspath(__file__))


class TFIDFTextEncoder(BaseEncoder):
    """Encode ``Document`` content from a `np.ndarray` (of strings) of length `BatchSize` into
    a `csr_matrix` of shape `Batchsize x EmbeddingDimension`.

    :param path_vectorizer: path containing the fitted tfidf encoder object
    :param args: not used
    :param kwargs: not used
    """

    def __init__(
        self,
        path_vectorizer: str = os.path.join(cur_dir, 'model/tfidf_vectorizer.pickle'),
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.path_vectorizer = path_vectorizer

    def post_init(self):
        import os
        import pickle

        super().post_init()
        if os.path.exists(self.path_vectorizer):
            self.tfidf_vectorizer = pickle.load(open(self.path_vectorizer, 'rb'))
        else:
            raise PretrainedModelFileDoesNotExist(
                f'{self.path_vectorizer} not found, cannot find a fitted tfidf_vectorizer'
            )

    def encode(self, content: np.ndarray, *args, **kwargs) -> 'scipy.sparse.csr_matrix':
        """Encode the ``Document`` content creating a tf-idf feature vector of the input.

        :param content: numpy array of strings containing the text data to be encoded
        :param args: not used
        :param kwargs: not used
        """
        return self.tfidf_vectorizer.transform(content)



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from os import path
from functools import wraps
from typing import Optional, Iterable, Tuple, Callable
from functools import lru_cache

import numpy as np
import copy

from jina.executors.decorators import as_update_method
from jina.executors.indexers.vector import NumpyIndexer, _ext_B, _euclidean, _norm, _cosine
from jina.helper import cached_property

if False:
    import zarr


def zarr_batching(
        func: Optional[Callable] = None,
        batch_size: Optional[int] = None,
):
    """Split the input of a function into small batches and call :func:`func` on each batch
    , collect the merged result and return. This is useful when the input is too big to fit into memory

    :param func: function to decorate
    :param batch_size: size of each batch
    :return: the merged result as if run :func:`func` once on the input.
    """

    def _zarr_batching(func):

        def batch_iterator(data, b_size):
            _l = data.shape[0]
            _d = data.ndim
            sl = [slice(None)] * _d
            for start in range(0, _l, b_size):
                end = min(_l, start + b_size)
                sl[0] = slice(start, end)
                _data_args = data[tuple(sl)]
                yield _data_args

        @wraps(func)
        def arg_wrapper(*args, **kwargs):
            # priority: decorator > class_attribute
            # by default data is in args[1] (self needs to be taken into account)
            data = args[2]
            b_size = batch_size or getattr(args[0], 'batch_size', None)

            # no batching if b_size is None
            if b_size is None or data is None:
                return func(*args, **kwargs)

            batch_args = list(copy.copy(args))

            results = []
            for _data_args in batch_iterator(data, b_size):
                batch_args[2] = _data_args
                r = func(*batch_args, **kwargs)
                if r is not None:
                    results.append(r)

            final_result = np.concatenate(results, 1)

            return final_result

        return arg_wrapper

    if func:
        return _zarr_batching(func)
    else:
        return _zarr_batching


class ZarrIndexer(NumpyIndexer):
    """
    Indexing based on Zarr arrays.

    For more information about Zarr, please check
    https://zarr.readthedocs.io/en/stable/index.html

    """

    @zarr_batching
    def _euclidean(self, cached_A, raw_B):
        data = _ext_B(raw_B)
        return _euclidean(cached_A, data)

    @zarr_batching
    def _cosine(self, cached_A, raw_B):
        data = _ext_B(_norm(raw_B))
        return _cosine(cached_A, data)

    @zarr_batching
    def _cdist(self, *args, **kwargs):
        with ImportExtensions(required=True):
            from scipy.spatial.distance import cdist
        return cdist(*args, **kwargs, metric=self.metric)

    def get_add_handler(self) -> 'zarr.hierarchy.Group':
        """
        Open an existing zarr group file for adding new vectors.

        :return: a zarr group file
        """
        import zarr
        return zarr.open(store=self.index_abspath, mode='a')

    def get_create_handler(self) -> 'zarr.hierarchy.Group':
        """
        Create a zarr group file for adding new vectors.

        :return: a zarr group file
        """
        import zarr
        return zarr.open(store=self.index_abspath, mode='w')

    @as_update_method
    def add(self, keys: Iterable[str], vectors: 'np.ndarray', *args, **kwargs) -> None:
        """
        Add the embeddings and document ids to the index.

        :param keys: a list of ``id``, i.e. ``doc.id`` in protobuf
        :param vectors: embeddings
        """
        np_keys = np.array(keys, (np.str_, self.key_length))

        self._add(np_keys, vectors)

    def _add(self, keys: 'np.ndarray', vectors: 'np.ndarray'):
        self._validate_key_vector_shapes(keys, vectors)
        if 'default' in self.write_handler.array_keys():
            self.write_handler['default'].append(data=vectors)
        else:
            self.write_handler.array(name='default', data=vectors)
        self.valid_indices = np.concatenate((self.valid_indices, np.full(len(keys), True)))
        self.key_bytes += keys.tobytes()
        self._size += keys.shape[0]

    @property
    def query_handler(self) -> Optional['zarr.core.Array']:
        """
        Get zarr file handler.

        :return: Zarr query handler
        """
        return self.get_query_handler()

    def get_query_handler(self) -> Optional['zarr.core.Array']:
        """
        Get zarr file handler.

        :return: Zarr query handler
        """
        import zarr
        if not (path.exists(self.index_abspath) or self.num_dim or self.dtype):
            return
        return zarr.open(store=f'{self.index_abspath}/default', mode='r',
                         shape=(self._size, self.num_dim), chunks=True)

    def query_by_key(self, keys: Iterable[str], *args, **kwargs) -> 'np.ndarray':
        """
        Get the vectors by keys.

        :param keys: list of document keys` as 1D-ndarray
        :return: subset of indexed vectors
        """
        filtered_keys = self._filter_nonexistent_keys(keys, self._ext2int_id.keys())
        int_ids = [self._ext2int_id[j] for j in filtered_keys]
        return self._raw_ndarray.get_orthogonal_selection(int_ids)

    @cached_property
    def _raw_ndarray(self):
        a = self.query_handler
        return a



from typing import Iterable, Optional, Dict

from jina.executors.indexers import BaseIndexer


class BaseQueryIndexer(BaseIndexer):
    """An indexer only for querying. It only reads once (at creation time, from a dump)"""

    def _post_init_wrapper(
        self,
        _metas: Optional[Dict] = None,
        _requests: Optional[Dict] = None,
        fill_in_metas: bool = True,
    ) -> None:
        super()._post_init_wrapper(_metas, _requests, fill_in_metas)
        self.dump_path = _metas.get('dump_path')
        # TODO this shouldn't be required
        # we don't do this for Compounds, as the _components
        # are not yet set at this stage.
        # for Compound we use a `_post_components`
        if self.dump_path and not hasattr(self, 'components'):
            self._load_dump(self.dump_path)

    def _load_dump(self, dump_path):
        """Load the dump at the dump_path

        :param dump_path: the path of the dump"""
        raise NotImplementedError

    def _log_warn(self):
        self.logger.error(f'Index {self.__class__} is write-once')

    def add(
        self, keys: Iterable[str], values: Iterable[bytes], *args, **kwargs
    ) -> None:
        """Disabled. QueryIndexers are write-once (at instantiation time)


        .. # noqa: DAR101
        """
        self._log_warn()

    def update(
        self, keys: Iterable[str], values: Iterable[bytes], *args, **kwargs
    ) -> None:
        """Disabled. QueryIndexers are write-once (at instantiation time)


        .. # noqa: DAR101
        """
        self._log_warn()

    def delete(self, keys: Iterable[str], *args, **kwargs) -> None:
        """Disabled. QueryIndexers are write-once (at instantiation time)


        .. # noqa: DAR101
        """
        self._log_warn()



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import inspect
from typing import Dict, Union, List

from .. import BaseExecutor
from ...helper import typename


class BaseCrafter(BaseExecutor):
    """
    A :class:`BaseCrafter` transforms the content of `Document`.
    It can be used for preprocessing, segmenting etc.
    It is an interface for Crafters which is a family of executors intended to apply
    transformations to single documents.
    The apply function is :func:`craft`, where the name of the arguments will be used as keys of the content.

    :param args: Additional positional arguments which are just used for the parent initialization
    :param kwargs: Additional keyword arguments which are just used for the parent initialization
    """

    def craft(self, *args, **kwargs) -> Union[List[Dict], Dict]:
        """
        Apply function of this executor.
        The name of the arguments are used as keys, which are then used to tell :class:`Driver` what information to extract
        from the protobuf request accordingly.
        The name of the arguments should be always valid keys defined in the protobuf.

        :param args: Extra variable length arguments
        :param kwargs: Extra variable keyword arguments
        """
        raise NotImplementedError



from typing import Optional, List

import numpy as np
from jina.executors.indexers import BaseIndexer


class BaseDBMSIndexer(BaseIndexer):
    """A class only meant for storing (indexing, update, delete) of data"""

    def add(
        self, ids: List[str], vecs: List[np.array], metas: List[bytes], *args, **kwargs
    ):
        """Add to the DBMS Indexer, both vectors and metadata

        :param ids: the ids of the documents
        :param vecs: the vectors
        :param metas: the metadata, in binary format
        :param args: not used
        :param kwargs: not used
        """
        raise NotImplementedError

    def update(
        self, ids: List[str], vecs: List[np.array], metas: List[bytes], *args, **kwargs
    ):
        """Update the DBMS Indexer, both vectors and metadata

        :param ids: the ids of the documents
        :param vecs: the vectors
        :param metas: the metadata, in binary format
        :param args: not used
        :param kwargs: not used
        """
        raise NotImplementedError

    def delete(self, ids: List[str], *args, **kwargs):
        """Delete from the indexer by ids

        :param ids: the ids of the Documents to delete
        :param args: not used
        :param kwargs: not used
        """
        raise NotImplementedError

    def dump(self, path: str, shards: int):
        """Dump the index

        :param path: the path to which to dump
        :param shards: the nr of shards to which to dump
        """
        raise NotImplementedError

    def query(self, key: str, *args, **kwargs) -> Optional[bytes]:
        """DBMSIndexers do NOT support querying

        :param key: the key by which to query
        :param args: not used
        :param kwargs: not used
        """
        raise NotImplementedError('DBMSIndexers do not support querying')



import pickle
from typing import List, Tuple, Generator
import numpy as np

from jina.executors.indexers.dump import export_dump_streaming
from jina.executors.indexers.dbms import BaseDBMSIndexer
from jina.executors.indexers.keyvalue import BinaryPbWriterMixin


class BinaryPbDBMSIndexer(BinaryPbWriterMixin, BaseDBMSIndexer):
    """A DBMS Indexer (no query method)"""

    def _get_generator(
        self, ids: List[str]
    ) -> Generator[Tuple[str, np.array, bytes], None, None]:
        for id_ in ids:
            vecs_metas_bytes = super()._query(id_)
            vec, meta = pickle.loads(vecs_metas_bytes)
            yield id_, vec, meta

    def dump(self, path: str, shards: int) -> None:
        """Dump the index

        :param path: the path to which to dump
        :param shards: the nr of shards to which to dump
        """
        self.write_handler.close()
        # noinspection PyPropertyAccess
        del self.write_handler
        self.handler_mutex = False
        ids = self.query_handler.header.keys()
        export_dump_streaming(
            path,
            shards=shards,
            size=self.size,
            data=self._get_generator(ids),
        )
        self.query_handler.close()
        self.handler_mutex = False
        # noinspection PyPropertyAccess
        del self.query_handler

    def add(
        self, ids: List[str], vecs: List[np.array], metas: List[bytes], *args, **kwargs
    ):
        """Add to the DBMS Indexer, both vectors and metadata

        :param ids: the ids of the documents
        :param vecs: the vectors
        :param metas: the metadata, in binary format
        :param args: not used
        :param kwargs: not used
        """
        if not any(ids):
            return

        vecs_metas = [pickle.dumps((vec, meta)) for vec, meta in zip(vecs, metas)]
        self._add(ids, vecs_metas)

    def update(
        self, ids: List[str], vecs: List[np.array], metas: List[bytes], *args, **kwargs
    ):
        """Update the DBMS Indexer, both vectors and metadata

        :param ids: the ids of the documents
        :param vecs: the vectors
        :param metas: the metadata, in binary format
        :param args: not used
        :param kwargs: not used
        """
        vecs_metas = [pickle.dumps((vec, meta)) for vec, meta in zip(vecs, metas)]
        keys, vecs_metas = self._filter_nonexistent_keys_values(
            ids, vecs_metas, self.query_handler.header.keys()
        )
        del self.query_handler
        self.handler_mutex = False
        if keys:
            self._delete(keys)
            self._add(keys, vecs_metas)

    def delete(self, ids: List[str], *args, **kwargs):
        """Delete the serialized documents from the index via document ids.

        :param ids: a list of ``id``, i.e. ``doc.id`` in protobuf
        :param args: not used
        :param kwargs: not used"""
        super(BinaryPbDBMSIndexer, self).delete(ids)


class KeyValueDBMSIndexer(BinaryPbDBMSIndexer):
    """An alias"""



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import mmap
import os
import random
from typing import Iterable, Optional

import numpy as np

from . import BaseKVIndexer
from ..compound import CompoundExecutor

HEADER_NONE_ENTRY = (-1, -1, -1)


class _WriteHandler:
    """
    Write file handler.

    :param path: Path of the file.
    :param mode: Writing mode. (e.g. 'ab', 'wb')
    """

    def __init__(self, path, mode):
        self.body = open(path, mode)
        self.header = open(path + '.head', mode)

    def close(self):
        """Close the file."""
        if not self.body.closed:
            self.body.close()
        if not self.header.closed:
            self.header.close()

    def flush(self):
        """Clear the body and header."""
        if not self.body.closed:
            self.body.flush()
        if not self.header.closed:
            self.header.flush()


class _ReadHandler:
    """
    Read file handler.

    :param path: Path of the file.
    :param key_length: Length of key.
    """

    def __init__(self, path, key_length):
        with open(path + '.head', 'rb') as fp:
            tmp = np.frombuffer(
                fp.read(),
                dtype=[
                    ('', (np.str_, key_length)),
                    ('', np.int64),
                    ('', np.int64),
                    ('', np.int64),
                ],
            )
            self.header = {
                r[0]: None
                if np.array_equal((r[1], r[2], r[3]), HEADER_NONE_ENTRY)
                else (r[1], r[2], r[3])
                for r in tmp
            }
        self._body = open(path, 'r+b')
        self.body = self._body.fileno()

    def close(self):
        """Close the file."""
        if not self._body.closed:
            self._body.close()


class BinaryPbWriterMixin:
    """Mixing for providing the binarypb writing and reading methods"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._start = 0
        self._page_size = mmap.ALLOCATIONGRANULARITY

    def get_add_handler(self) -> '_WriteHandler':
        """
        Get write file handler.

        :return: write handler
        """
        # keep _start position as in pickle serialization
        return _WriteHandler(self.index_abspath, 'ab')

    def get_create_handler(self) -> '_WriteHandler':
        """
        Get write file handler.

        :return: write handler.
        """
        self._start = 0  # override _start position
        return _WriteHandler(self.index_abspath, 'wb')

    def get_query_handler(self) -> '_ReadHandler':
        """
        Get read file handler.

        :return: read handler.
        """
        return _ReadHandler(self.index_abspath, self.key_length)

    def _add(
        self, keys: Iterable[str], values: Iterable[bytes], writer: _WriteHandler = None
    ):
        if writer is None:
            if getattr(self, 'write_handler'):
                writer = self.write_handler
            else:
                raise Exception(
                    'No writer provided to add() and no self.write_handler available'
                )

        for key, value in zip(keys, values):
            l = len(value)  #: the length
            p = (
                int(self._start / self._page_size) * self._page_size
            )  #: offset of the page
            r = (
                self._start % self._page_size
            )  #: the remainder, i.e. the start position given the offset
            # noinspection PyTypeChecker
            writer.header.write(
                np.array(
                    (key, p, r, r + l),
                    dtype=[
                        ('', (np.str_, self.key_length)),
                        ('', np.int64),
                        ('', np.int64),
                        ('', np.int64),
                    ],
                ).tobytes()
            )
            self._start += l
            writer.body.write(value)
            self._size += 1
        writer.flush()

    def delete(self, keys: Iterable[str], *args, **kwargs) -> None:
        """Delete the serialized documents from the index via document ids.

        :param keys: a list of ``id``, i.e. ``doc.id`` in protobuf
        :param args: not used
        :param kwargs: not used
        """
        keys = self._filter_nonexistent_keys(keys, self.query_handler.header.keys())
        del self.query_handler
        self.handler_mutex = False
        if keys:
            self._delete(keys)

    def _delete(self, keys: Iterable[str]) -> None:
        for key in keys:
            self.write_handler.header.write(
                np.array(
                    tuple(np.concatenate([[key], HEADER_NONE_ENTRY])),
                    dtype=[
                        ('', (np.str_, self.key_length)),
                        ('', np.int64),
                        ('', np.int64),
                        ('', np.int64),
                    ],
                ).tobytes()
            )
            self._size -= 1

    def _query(self, key):
        pos_info = self.query_handler.header.get(key, None)
        if pos_info is not None:
            p, r, l = pos_info
            with mmap.mmap(self.query_handler.body, offset=p, length=l) as m:
                return m[r:]


class BinaryPbIndexer(BinaryPbWriterMixin, BaseKVIndexer):
    """Simple Key-value indexer."""

    def __getstate__(self):
        # called on pickle save
        if self.delete_on_dump:
            self._delete_invalid_indices()
        d = super().__getstate__()
        return d

    def _delete_invalid_indices(self):
        if self.query_handler:
            self.query_handler.close()
        if self.write_handler:
            self.write_handler.flush()
            self.write_handler.close()

        keys = []
        vals = []
        # we read the valid values and write them to the intermediary file
        read_handler = _ReadHandler(self.index_abspath, self.key_length)
        for key in read_handler.header.keys():
            pos_info = read_handler.header.get(key, None)
            if pos_info:
                p, r, l = pos_info
                with mmap.mmap(read_handler.body, offset=p, length=l) as m:
                    keys.append(key)
                    vals.append(m[r:])
        read_handler.close()
        if len(keys) == 0:
            return

        # intermediary file
        tmp_file = self.index_abspath + '-tmp'
        self._start = 0
        filtered_data_writer = _WriteHandler(tmp_file, 'ab')
        # reset size
        self._size = 0
        self._add(keys, vals, filtered_data_writer)
        filtered_data_writer.close()

        # replace orig. file
        # and .head file
        head_path = self.index_abspath + '.head'
        os.remove(self.index_abspath)
        os.remove(head_path)
        os.rename(tmp_file, self.index_abspath)
        os.rename(tmp_file + '.head', head_path)

    def __init__(self, delete_on_dump: bool = False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.delete_on_dump = delete_on_dump

    def add(
        self, keys: Iterable[str], values: Iterable[bytes], *args, **kwargs
    ) -> None:
        """Add the serialized documents to the index via document ids.

        :param keys: a list of ``id``, i.e. ``doc.id`` in protobuf
        :param values: serialized documents
        :param args: extra arguments
        :param kwargs: keyword arguments
        """
        if not any(keys):
            return

        self._add(keys, values)

    def sample(self) -> Optional[bytes]:
        """Return a random entry from the indexer for sanity check.

        :return: A random entry from the indexer.
        """
        k = random.sample(self.query_handler.header.keys(), k=1)[0]
        return self[k]

    def __iter__(self):
        for k in self.query_handler.header.keys():
            yield self[k]

    def query(self, key: str, *args, **kwargs) -> Optional[bytes]:
        """Find the serialized document to the index via document id.

        :param key: document id
        :param args: extra arguments
        :param kwargs: keyword arguments
        :return: serialized documents
        """
        return self._query(key)

    def update(
        self, keys: Iterable[str], values: Iterable[bytes], *args, **kwargs
    ) -> None:
        """Update the serialized documents on the index via document ids.

        :param keys: a list of ``id``, i.e. ``doc.id`` in protobuf
        :param values: serialized documents
        :param args: extra arguments
        :param kwargs: keyword arguments
        """
        keys, values = self._filter_nonexistent_keys_values(
            keys, values, self.query_handler.header.keys()
        )
        del self.query_handler
        self.handler_mutex = False
        if keys:
            self._delete(keys)
            self.add(keys, values)

    def delete(self, keys: Iterable[str], *args, **kwargs) -> None:
        """Delete the serialized documents from the index via document ids.

        :param keys: a list of ``id``, i.e. ``doc.id`` in protobuf
        :param args: not used
        :param kwargs: not used"""
        super(BinaryPbIndexer, self).delete(keys)


class KeyValueIndexer(BinaryPbIndexer):
    """Alias for :class:`BinaryPbIndexer` """


class DataURIPbIndexer(BinaryPbIndexer):
    """Alias for BinaryPbIndexer"""


class UniquePbIndexer(CompoundExecutor):
    """A frequently used pattern for combining a :class:`BaseKVIndexer` and a :class:`DocCache` """



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Dict

import numpy as np

from jina.executors.decorators import single
from jina.executors.crafters import BaseCrafter


class ArrayBytesReader(BaseCrafter):
    """
    Convert a byte stream into a numpy array and save to the Document.

    The size of the vectors is provided in the constructor
    so that the numpy array can be interpreted properly.

    :param as_type: The numpy array will be this type
    :param args:  Additional positional arguments
    :param kwargs: Additional keyword arguments
    """

    def __init__(self, as_type: str = 'float32', *args, **kwargs):
        """Set constructor"""
        super().__init__(*args, **kwargs)
        self.as_type = as_type

    @single
    def craft(self, buffer: bytes, *args, **kwargs) -> Dict:
        """
        Split string into numbers and convert to numpy array.

        :param buffer: the bytes representing the array
        :param args:  Additional positional arguments
        :param kwargs: Additional keyword arguments
        :return: a chunk dict with the numpy array
        """
        _array = np.frombuffer(buffer, self.as_type)
        return dict(blob=_array)



from jina.executors.compound import CompoundExecutor
from jina.executors.indexers.query import BaseQueryIndexer


class CompoundQueryExecutor(CompoundExecutor, BaseQueryIndexer):
    """A Compound Executor that wraps several QueryIndexers

    :param dump_path: the path to initialize from
    """

    # TODO this shouldn't be required
    # we don't do this for Compounds, as the _components
    # are not yet set at this stage.
    # for Compound we use a `_post_components`
    def _post_components(self):
        self._load_dump(self.dump_path)

    def _load_dump(self, dump_path, *args, **kwargs):
        """Loads the data in the indexer

        :param dump_path: the path to the dump
        :param args: passed to the inner Indexer's load_dump
        :param kwargs: passed to the inner Indexer's load_dump
        """
        for c in self.components:
            c._load_dump(dump_path)


class CompoundQueryIndexer(CompoundQueryExecutor):
    """Alias"""



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from .peas import BasePea as Pea
from .pods import BasePod
from .pods import Pod
from .pods.compoundpod import CompoundPod



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import os
from typing import Optional

import numpy as np

from jina.executors.decorators import batching, as_ndarray
from jina.executors.encoders import BaseAudioEncoder
from jina.executors.encoders.frameworks import BaseTorchEncoder
from jina.excepts import PretrainedModelFileDoesNotExist
from jina.helper import cached_property


class Wav2VecSpeechEncoder(BaseTorchEncoder, BaseAudioEncoder):
    """
    Use a pre-trained model (`wav2vec`) to encode audio signal.

    :class:`Wav2VecSpeechEncoder` is a speech encoder based on `wav2vec`,
        an unsupervised pre-trained model for speech recognition presented and implemented
        by Facebook: https://github.com/pytorch/fairseq/tree/master/examples/wav2vec
        It uses a pre-trained model to encode an audio signal from
        a `Batch x Signal Length` ndarray into a `Batch x Concatenated Features` ndarray,
        and produces a representation for each time step at a rate of 100 Hz.

    :param model_path: the path of the pre-trained model.
        The pre-trained model can be downloaded at
        https://github.com/pytorch/fairseq/tree/master/examples/wav2vec/README.md#wav2vec
    :param input_sample_rate: input sampling rate in Hz (22050 by default)
    """

    def __init__(self,
                 model_path: Optional[str] = '/tmp/wav2vec_large.pt',
                 input_sample_rate: int = 22050,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.model_path = model_path
        self.input_sample_rate = input_sample_rate

    def post_init(self):
        """Load Wav2Vec model"""
        super().post_init()
        if self.model_path and os.path.exists(self.model_path):
            import torch
            from fairseq.models.wav2vec import Wav2VecModel
            cp = torch.load(self.model_path, map_location=torch.device('cpu'))
            self.model = Wav2VecModel.build_model(cp['args'], task=None)
            self.model.load_state_dict(cp['model'])
            self.model.eval()
            self.to_device(self.model)
            self._tensor_func = torch.tensor
        else:
            raise PretrainedModelFileDoesNotExist(f'model at {self.model_path} does not exist')

    @batching
    @as_ndarray
    def encode(self, content: np.ndarray, *args, **kwargs) -> np.ndarray:
        """
        Resample  input audio signal to 16kHz.

        Segments the resampled signal of each Doc into `wav2vec` frames,
        encodes the frames and concatenates Doc frame embeddings into a
        single Doc embedding.

        :param content: A`Batch x Signal Length` ndarray, where
            `Signal Length` is a number of samples
        :return: A `Batch x Concatenated Features` ndarray,
            where `Concatenated Features` is a 512-dimensional feature
            vector times the number of the wav2vec frames.
        """
        assert content.shape[1] >= 465, 'the signal must have at least 465 samples'
        from librosa import resample
        embeds = []
        with self.session():
            for chunk_data in content:
                resampled_signal = resample(chunk_data, self.input_sample_rate, 16000)
                signal_tensor = self.array2tensor(resampled_signal.reshape(1, -1))
                features = self.model.feature_extractor(signal_tensor)
                embed_tensor = self.model.feature_aggregator(features)[0]
                chunk_embed = self.tensor2array(embed_tensor).T.flatten()
                embeds.append(chunk_embed)
        return embeds

    def array2tensor(self, array):
        """Transform array into tensor"""
        tensor = self._tensor_func(array)
        return tensor.cuda() if self.on_gpu else tensor

    def tensor2array(self, tensor):
        """Transform tensor into array"""
        return tensor.cuda().numpy() if self.on_gpu else tensor.numpy()

    @cached_property
    def session(self):
        return self.get_session()

    def get_session(self):
        """Get no_grad from torch"""
        from torch import no_grad
        return no_grad


"""Decorators and wrappers designed for wrapping :class:`BaseExecutor` functions. """

__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import copy
import inspect
from functools import wraps
from itertools import islice, chain
from typing import Callable, Any, Union, Iterator, List, Optional, Dict, Iterable

import numpy as np

from .metas import get_default_metas
from ..helper import batch_iterator, convert_tuple_to_list
from ..logging import default_logger


def as_aggregate_method(func: Callable) -> Callable:
    """Mark a function so that it keeps track of the number of documents evaluated and a running sum
    to have always access to average value
    :param func: the function to decorate
    :return: the wrapped function
    """

    @wraps(func)
    def arg_wrapper(self, *args, **kwargs):
        f = func(self, *args, **kwargs)
        self._running_stats += f
        return f

    return arg_wrapper


def as_update_method(func: Callable) -> Callable:
    """Mark the function as the updating function of this executor,
    calling this function will change the executor so later you can save the change via :func:`save`
    Will set the is_updated property after function is called.
    :param func: the function to decorate
    :return: the wrapped function
    """

    @wraps(func)
    def arg_wrapper(self, *args, **kwargs):
        f = func(self, *args, **kwargs)
        self.is_updated = True
        return f

    return arg_wrapper


def wrap_func(cls, func_lst, wrapper):
    """Wrapping a class method only once, inherited but not overridden method will not be wrapped again

    :param cls: class
    :param func_lst: function list to wrap
    :param wrapper: the wrapper
    """
    for f_name in func_lst:
        if hasattr(cls, f_name) and all(
            getattr(cls, f_name) != getattr(i, f_name, None) for i in cls.mro()[1:]
        ):
            setattr(cls, f_name, wrapper(getattr(cls, f_name)))


def as_ndarray(func: Callable, dtype=np.float32) -> Callable:
    """Convert an :class:`BaseExecutor` function returns to a ``numpy.ndarray``,
    the following type are supported: `EagerTensor`, `Tensor`, `list`

    :param func: the function to decorate
    :param dtype: the converted dtype of the ``numpy.ndarray``
    :return: the wrapped function
    """

    @wraps(func)
    def arg_wrapper(self, *args, **kwargs):
        r = func(self, *args, **kwargs)
        r_type = type(r).__name__
        if r_type in {'ndarray', 'EagerTensor', 'Tensor', 'list'}:
            return np.array(r, dtype)
        else:
            raise TypeError(f'unrecognized type {r_type}: {type(r)}')

    return arg_wrapper


def store_init_kwargs(func: Callable) -> Callable:
    """Mark the args and kwargs of :func:`__init__` later to be stored via :func:`save_config` in YAML
    :param func: the function to decorate
    :return: the wrapped function
    """

    @wraps(func)
    def arg_wrapper(self, *args, **kwargs):
        if func.__name__ != '__init__':
            raise TypeError(
                'this decorator should only be used on __init__ method of an executor'
            )
        taboo = {'self', 'args', 'kwargs'}
        _defaults = get_default_metas()
        taboo.update(_defaults.keys())
        all_pars = inspect.signature(func).parameters
        tmp = {k: v.default for k, v in all_pars.items() if k not in taboo}
        tmp_list = [k for k in all_pars.keys() if k not in taboo]
        # set args by aligning tmp_list with arg values
        for k, v in zip(tmp_list, args):
            tmp[k] = v
        # set kwargs
        for k, v in kwargs.items():
            if k in tmp:
                tmp[k] = v

        if getattr(self, 'store_args_kwargs', None):
            if args:
                tmp['args'] = args
            if kwargs:
                tmp['kwargs'] = {k: v for k, v in kwargs.items() if k not in taboo}

        if hasattr(self, '_init_kwargs_dict'):
            self._init_kwargs_dict.update(tmp)
        else:
            self._init_kwargs_dict = tmp
        convert_tuple_to_list(self._init_kwargs_dict)
        f = func(self, *args, **kwargs)
        return f

    return arg_wrapper


def _get_slice(
    data: Union[Iterator[Any], List[Any], np.ndarray], total_size: int
) -> Union[Iterator[Any], List[Any], np.ndarray]:
    if isinstance(data, Dict):
        data = islice(data.items(), total_size)
    else:
        data = data[:total_size]
    return data


def _get_size(data: Union[Iterator[Any], List[Any], np.ndarray], axis: int = 0) -> int:
    if isinstance(data, np.ndarray):
        total_size = data.shape[axis]
    elif hasattr(data, '__len__'):
        total_size = len(data)
    else:
        total_size = None
    return total_size


def _get_total_size(full_data_size, batch_size, num_batch):
    batched_data_size = batch_size * num_batch if num_batch else None

    if full_data_size is not None and batched_data_size is not None:
        total_size = min(full_data_size, batched_data_size)
    else:
        total_size = full_data_size or batched_data_size
    return total_size


def _merge_results_after_batching(
    final_result, merge_over_axis: int = 0, flatten: bool = True
):
    if not final_result:
        return

    if isinstance(final_result[0], np.ndarray):
        if len(final_result[0].shape) > 1:
            final_result = np.concatenate(final_result, merge_over_axis)
    elif isinstance(final_result[0], list) and flatten:
        final_result = list(chain.from_iterable(final_result))

    return final_result


def batching(
    func: Optional[Callable[[Any], np.ndarray]] = None,
    batch_size: Optional[Union[int, Callable]] = None,
    num_batch: Optional[int] = None,
    split_over_axis: int = 0,
    merge_over_axis: int = 0,
    slice_on: int = 1,
    slice_nargs: int = 1,
    label_on: Optional[int] = None,
    ordinal_idx_arg: Optional[int] = None,
    flatten_output: bool = True,
) -> Any:
    """Split the input of a function into small batches and call :func:`func` on each batch
    , collect the merged result and return. This is useful when the input is too big to fit into memory

    :param func: function to decorate
    :param batch_size: size of each batch
    :param num_batch: number of batches to take, the rest will be ignored
    :param split_over_axis: split over which axis into batches
    :param merge_over_axis: merge over which axis into a single result
    :param slice_on: the location of the data. When using inside a class,
            ``slice_on`` should take ``self`` into consideration.
    :param slice_nargs: the number of arguments
    :param label_on: the location of the labels. Useful for data with any kind of accompanying labels
    :param ordinal_idx_arg: the location of the ordinal indexes argument. Needed for classes
            where function decorated needs to know the ordinal indexes of the data in the batch
            (Not used when label_on is used)
    :param flatten_output: If this is set to True, the results from different batches will be chained and the returning value is a list of the results. Otherwise, the returning value is a list of lists, in which each element is a list containing the result from one single batch. Note if there is only one batch returned, the returned result is always flatten.
    :return: the merged result as if run :func:`func` once on the input.

    Example:
        .. highlight:: python
        .. code-block:: python

            class MemoryHungryExecutor:

                @batching
                def train(self, batch: 'numpy.ndarray', *args, **kwargs):
                    gpu_train(batch)  #: this will respect the ``batch_size`` defined as object attribute

                @batching(batch_size = 64)
                def train(self, batch: 'numpy.ndarray', *args, **kwargs):
                    gpu_train(batch)
    """

    def _batching(func):
        @wraps(func)
        def arg_wrapper(*args, **kwargs):
            # priority: decorator > class_attribute
            # by default data is in args[1] (self needs to be taken into account)
            data = args[slice_on : slice_on + slice_nargs]
            b_size = (
                batch_size(data) if callable(batch_size) else batch_size
            ) or getattr(args[0], 'batch_size', None)

            # no batching if b_size is None
            if b_size is None or data is None:
                return func(*args, **kwargs)

            default_logger.debug(
                f'batching enabled for {func.__qualname__} batch_size={b_size} '
                f'num_batch={num_batch} axis={split_over_axis}'
            )

            results = []
            data = (data, args[label_on]) if label_on else data

            yield_slice = [
                isinstance(args[slice_on + i], np.memmap) for i in range(0, slice_nargs)
            ]

            slice_idx = None

            # split the data into batches
            data_iterators = [
                batch_iterator(
                    data[i],
                    b_size,
                    split_over_axis,
                    yield_slice=yield_slice[i],
                )
                for i in range(0, slice_nargs)
            ]

            batch_args = list(copy.copy(args))

            # load the batches of data and feed into the function
            for _data_args in zip(*data_iterators):
                _data_args = list(_data_args)
                for i, (_yield_slice, _arg) in enumerate(zip(yield_slice, _data_args)):
                    if _yield_slice:
                        original_arg = args[slice_on + i]
                        _memmap = np.memmap(
                            original_arg.filename,
                            dtype=original_arg.dtype,
                            mode='r',
                            shape=original_arg.shape,
                        )
                        _data_args[i] = _memmap[_arg]
                        slice_idx = _arg[split_over_axis]
                        if slice_idx.start is None or slice_idx.stop is None:
                            slice_idx = None
                        del _memmap

                    # TODO: figure out what is ordinal_idx_arg
                    if not isinstance(_data_args[i], tuple):
                        if ordinal_idx_arg and slice_idx is not None:
                            batch_args[ordinal_idx_arg] = slice_idx

                batch_args[slice_on : slice_on + slice_nargs] = _data_args

                r = func(*batch_args, **kwargs)

                if r is not None:
                    results.append(r)

            return _merge_results_after_batching(
                results, merge_over_axis, flatten_output
            )

        return arg_wrapper

    if func:
        return _batching(func)
    else:
        return _batching


def single(
    func: Optional[Callable[[Any], np.ndarray]] = None,
    merge_over_axis: int = 0,
    slice_on: int = 1,
    slice_nargs: int = 1,
    flatten_output: bool = False,
) -> Any:
    """Guarantee that the inputs of a function with more than one argument is provided as single instances and not in batches

    :param func: function to decorate
    :param merge_over_axis: merge over which axis into a single result
    :param slice_on: the location of the data. When using inside a class,
            ``slice_on`` should take ``self`` into consideration.
    :param slice_nargs: the number of positional arguments considered as data
    :param flatten_output: If this is set to True, the results from different batches will be chained and the returning value is a list of the results. Otherwise, the returning value is a list of lists, in which each element is a list containing the result from one single batch. Note if there is only one batch returned, the returned result is always flatten.
    :return: the merged result as if run :func:`func` once on the input.

    ..warning:
        data arguments will be taken starting from ``slice_on` to ``slice_on + num_data``

    Example:
        .. highlight:: python
        .. code-block:: python

            class OneByOneCrafter:

                @single
                def craft(self, text: str, id: str) -> Dict:
            ...

    .. note:
        Single multi input decorator will let the user interact with the executor in 3 different ways:
            - Providing batches: (This decorator will make sure that the actual method receives just a single instance)
            - Providing a single instance
            - Providing a single instance through kwargs.

        .. highlight:: python
        .. code-block:: python

            class OneByOneCrafter:
                @single
                def craft(self, text: str, id: str) -> Dict:
                    return {'text': f'{text}-crafted', 'id': f'{id}-crafted'}

            crafter = OneByOneCrafter()

            results = crafted.craft(['text1', 'text2'], ['id1', 'id2'])
            assert len(results) == 2
            assert results[0] == {'text': 'text1-crafted', 'id': 'id1-crafted'}
            assert results[1] == {'text': 'text2-crafted', 'id': 'id2-crafted'}

            result = crafter.craft('text', 'id')
            assert result['text'] == 'text-crafted'
            assert result['id'] == 'id-crafted'

            results = crafted.craft(text='text', id='id')
            assert result['text'] == 'text-crafted'
            assert result['id'] == 'id-crafted'
    """

    def _single_multi_input(func):
        @wraps(func)
        def arg_wrapper(*args, **kwargs):
            # by default data is in args[1:] (self needs to be taken into account)
            args = list(args)
            default_logger.debug(f'batching disabled for {func.__qualname__}')

            data_iterators = args[slice_on : slice_on + slice_nargs]

            if len(args) <= slice_on:
                # like this one can use the function with single kwargs
                return func(*args, **kwargs)
            elif len(args) < slice_on + slice_nargs:
                raise IndexError(
                    f'can not select positional args at {slice_on}: {slice_nargs}, '
                    f'your `args` has {len(args)} arguments.'
                )
            elif (
                len(args) <= slice_on
                or isinstance(data_iterators[0], str)
                or isinstance(data_iterators[0], bytes)
                or not isinstance(data_iterators[0], Iterable)
            ):
                # like this one can use the function with single kwargs
                return func(*args, **kwargs)

            final_result = []
            for new_args in zip(*data_iterators):
                args[slice_on : slice_on + slice_nargs] = new_args
                r = func(*args, **kwargs)

                if r is not None:
                    final_result.append(r)

            return _merge_results_after_batching(
                final_result, merge_over_axis, flatten=flatten_output
            )

        return arg_wrapper

    if func:
        return _single_multi_input(func)
    else:
        return _single_multi_input


def requests(func: Callable = None, on: str = 'default') -> Callable:
    """Decorator for binding an Executor function to requests

    :param func: the Executor function to decorate
    :param on: the request type to bind, e.g. IndexRequest, SearchRequest, UpdateRequest, DeleteRequest, etc.
            you may also use `index`, `search`, `update`, `delete` as shortcut.
    :return: the wrapped function
    """

    def _requests(func):
        @wraps(func)
        def arg_wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return arg_wrapper

    if func:
        return _requests(func)
    else:
        return _requests



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Optional

from . import FlatRecursiveMixin, BaseExecutableDriver, DocsExtractUpdateMixin


class CraftDriver(DocsExtractUpdateMixin, FlatRecursiveMixin, BaseExecutableDriver):
    """Drivers inherited from this Driver will bind :meth:`craft` by default """

    def __init__(
        self, executor: Optional[str] = None, method: str = 'craft', *args, **kwargs
    ):
        super().__init__(executor, method, *args, **kwargs)



"""Argparser module for Pea runtimes"""
import argparse

from ..helper import add_arg_group, _SHOW_ALL_ARGS, KVAppendAction
from ...enums import PeaRoleType, RuntimeBackendType
from ...peapods.runtimes import list_all_runtimes


def mixin_pea_parser(parser):
    """Mixing in arguments required by :class:`BasePea` into the given parser.
    :param parser: the parser instance to which we add arguments
    """

    gp = add_arg_group(parser, title='Pea')

    gp.add_argument(
        '--daemon',
        action='store_true',
        default=False,
        help='The Pea attempts to terminate all of its Runtime child processes/threads on existing. '
        'setting it to true basically tell the Pea do not wait on the Runtime when closing',
    )

    gp.add_argument(
        '--runtime-backend',
        '--runtime',
        type=RuntimeBackendType.from_string,
        choices=list(RuntimeBackendType),
        default=RuntimeBackendType.PROCESS,
        help='The parallel backend of the runtime inside the Pea',
    )

    gp.add_argument(
        '--runtime-cls',
        type=str,
        choices=list_all_runtimes(),
        default='ZEDRuntime',
        help='The runtime class to run inside the Pea',
    )

    gp.add_argument(
        '--timeout-ready',
        type=int,
        default=60000,
        help='The timeout in milliseconds of a Pea waits for the runtime to be ready, -1 for waiting '
        'forever',
    )

    gp.add_argument(
        '--env',
        action=KVAppendAction,
        metavar='KEY=VALUE',
        nargs='*',
        help='The map of environment variables that are available inside runtime',
    )

    gp.add_argument(
        '--expose-public',
        action='store_true',
        default=False,
        help='If set, expose the public IP address to remote when necessary, by default it exposes'
        'private IP address, which only allows accessing under the same network/subnet. Important to '
        'set this to true when the Pea will receive input connections from remote Peas',
    )

    # hidden CLI used for internal only

    gp.add_argument(
        '--pea-id',
        type=int,
        default=0,
        help='defines the suffix for the workspace path of the pea`'
        if _SHOW_ALL_ARGS
        else argparse.SUPPRESS,
    )

    gp.add_argument(
        '--pea-role',
        type=PeaRoleType.from_string,
        choices=list(PeaRoleType),
        default=PeaRoleType.SINGLETON,
        help='The role of this Pea in a Pod' if _SHOW_ALL_ARGS else argparse.SUPPRESS,
    )

    gp.add_argument(
        '--noblock-on-start',
        action='store_true',
        default=False,
        help='If set, starting a Pea/Pod does not block the thread/process. It then relies on '
        '`wait_start_success` at outer function for the postpone check.'
        if _SHOW_ALL_ARGS
        else argparse.SUPPRESS,
    )



import numpy as np

from jina.executors.decorators import batching, as_ndarray
from jina.executors.encoders.frameworks import BaseTorchEncoder

class FarmTextEncoder(BaseTorchEncoder):
    """
    Encode an array of string in size `B` into an ndarray in size `B x D`

    The ndarray potentially is BatchSize x (Channel x Height x Width)
    into an ndarray in size `B x D`. Where `B` is the batch size and
    `D` is the Dimension.

    FARM-based text encoder: (Framework for Adapting Representation Models)
    https://github.com/deepset-ai/FARM

    :param model_name_or_path: Directory or public name of the model to load.
    :param num_processes: Number of processes for `multiprocessing.Pool`.
        Set to value of 0 to disable multiprocessing. Set to None to let
        Inferencer use all CPU cores. If you want to debug the Language Model,
        you might need to disable multiprocessing
    :param extraction_strategy: Strategy to extract vectors. Choices:
        'cls_token' (sentence vector), 'reduce_mean'(sentence vector),
        reduce_max (sentence vector), 'per_token' (individual token vectors)
    :param extraction_layer: number of layer from which the embeddings shall
        be extracted. Default: -1 (very last layer).
    :param args:  Additional positional arguments
    :param kwargs: Additional keyword arguments
    """

    def __init__(self, pretrained_model_name_or_path: str = 'deepset/bert-base-cased-squad2',
                 num_processes: int = 0, extraction_strategy: str = 'cls_token',
                 extraction_layer: int = -1,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.pretrained_model_name_or_path = pretrained_model_name_or_path
        self.num_processes = num_processes
        self.extraction_strategy = extraction_strategy
        self.extraction_layer = extraction_layer

    def post_init(self):
        """Load FARM-based text model"""
        from farm.infer import Inferencer
        self.model = Inferencer.load(model_name_or_path=self.pretrained_model_name_or_path, task_type='embeddings',
                                     num_processes=self.num_processes)

    @batching
    @as_ndarray
    def encode(self, content: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        """
        Encode an array of string in size `B` into an ndarray in size `B x D`.

        The ndarray potentially is BatchSize x (Channel x Height x Width)
            into an ndarray in size `B x D`. Where `D` is the Dimension.

        :param content: a 1d array of string type in size `B`
        :return: an ndarray in size `B x D`.
        """
        basic_texts = [{'text': s} for s in content]
        embeds = np.stack([k['vec'] for k in self.model.extract_vectors(dicts=basic_texts)])
        return embeds


__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Optional

from jina.logging.logger import JinaLogger

class PostgreSQLDBMSHandler:
    """
    Postgres Handler to connect to the database and can apply add, update, delete and query.

    :param hostname: hostname of the machine
    :param port: the port
    :param username: the username to authenticate
    :param password: the password to authenticate
    :param database: the database name
    :param collection: the collection name
    :param args: other arguments
    :param kwargs: other keyword arguments
    """

    def __init__(self,
                 hostname: str = '127.0.0.1',
                 port: int = 5432,
                 username: str = 'default_name',
                 password: str = 'default_pwd',
                 database: str = 'postgres',
                 table: Optional[str] = 'default_table',
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = JinaLogger(self.__class__.__name__)
        self.hostname = hostname
        self.port = port
        self.username = username
        self.password = password
        self.database = database
        self.table = table

    def __enter__(self):
        return self.connect()

    def connect(self) -> 'PostgreSQLDBMSHandler':
        """Connect to the database. """

        import psycopg2
        from psycopg2 import Error

        try:
            self.connection = psycopg2.connect(user=self.username,
                password=self.password,
                database=self.database,
                host=self.hostname,
                port=self.port)
            self.cursor = self.connection.cursor()
            self.logger.info('Successfully connected to the database')
            self.use_table()
            self.connection.commit()
        except (Exception, Error) as error:
            self.logger.error('Error while connecting to PostgreSQL', error)
        return self

    def use_table(self):
        """
        Use table if exists or create one if it doesn't.

        Create table if needed with id, vecs and metas.
        """
        from psycopg2 import Error

        self.cursor.execute('select exists(select * from information_schema.tables where table_name=%s)', (self.table,))
        if self.cursor.fetchone()[0]:
            self.logger.info('Using existing table')
        else:
            try:
                self.cursor.execute(
                   f"CREATE TABLE {self.table} ( \
                    ID VARCHAR PRIMARY KEY,  \
                    VECS BYTEA,  \
                    METAS BYTEA);"
                )
                self.logger.info('Successfully created table')
            except (Exception, Error) as error:
                self.logger.error('Error while creating table!')

    def add(self, ids, vecs, metas, *args, **kwargs):
        """ Insert the documents into the database.

        :param ids: List of doc ids to be added
        :param vecs: List of vecs to be added
        :param metas: List of metas of docs to be added
        :param args: other arguments
        :param kwargs: other keyword arguments
        :param args: other arguments
        :param kwargs: other keyword arguments
        :return record: List of Document's id added
        """
        row_count = 0
        for i in range(len(ids)):
            self.cursor.execute(
                f'INSERT INTO {self.table} (ID, VECS, METAS) VALUES (%s, %s, %s)',
                (ids[i], vecs[i].tobytes(), metas[i]),
            )
            row_count += self.cursor.rowcount
        self.connection.commit()
        return row_count

    def update(self, ids, vecs, metas, *args, **kwargs):
        """ Updated documents from the database.

        :param ids: Ids of Doc to be updated
        :param vecs: List of vecs to be updated
        :param metas: List of metas of docs to be updated
        :param args: other arguments
        :param kwargs: other keyword arguments
        :return record: List of Document's id after update
        """
        row_count = 0

        for i in range(len(ids)):
            self.cursor.execute(
                f'UPDATE {self.table} SET VECS = %s, METAS = %s WHERE ID = %s',
                (vecs[i].tobytes(), metas[i], ids[i]),
            )
            row_count += self.cursor.rowcount
        self.connection.commit()
        return row_count

    def delete(self, ids, *args, **kwargs):
        """ Delete document from the database.

        :param ids: ids of Documents to be removed
        :param args: other arguments
        :param kwargs: other keyword arguments
        :return record: List of Document's id after deletion
         """
        row_count = 0
        for id in ids:
            self.cursor.execute(f'DELETE FROM {self.table} where (ID) = (%s);', (id,))
            row_count += self.cursor.rowcount
        self.connection.commit()
        return row_count

    def __exit__(self, *args):
        """ Make sure the connection to the database is closed."""

        from psycopg2 import Error
        try:
            self.connection.close()
            self.cursor.close()
            self.logger.info('PostgreSQL connection is closed')
        except (Exception, Error) as error:
            self.logger.error('Error while closing: ', error)



import os
import sys
from typing import Tuple, Generator, BinaryIO, TextIO

import numpy as np

BYTE_PADDING = 4


def export_dump_streaming(
    path: str,
    shards: int,
    size: int,
    data: Generator[Tuple[str, np.array, bytes], None, None],
):
    """Export the data to a path, based on sharding,

    :param path: path to dump
    :param shards: the nr of shards this pea is part of
    :param size: total amount of entries
    :param data: the generator of the data (ids, vectors, metadata)
    """
    _handle_dump(data, path, shards, size)


def _handle_dump(
    data: Generator[Tuple[str, np.array, bytes], None, None],
    path: str,
    shards: int,
    size: int,
):
    if os.path.exists(path):
        raise Exception(f'path for dump {path} already exists. Not dumping...')
    size_per_shard = size // shards
    extra = size % shards
    shard_range = list(range(shards))
    for shard_id in shard_range:
        if shard_id == shard_range[-1]:
            size_this_shard = size_per_shard + extra
        else:
            size_this_shard = size_per_shard
        _write_shard_data(data, path, shard_id, size_this_shard)


def _write_shard_data(
    data: Generator[Tuple[str, np.array, bytes], None, None],
    path: str,
    shard_id: int,
    size_this_shard: int,
):
    shard_path = os.path.join(path, str(shard_id))
    shard_docs_written = 0
    os.makedirs(shard_path)
    vectors_fp, metas_fp, ids_fp = _get_file_paths(shard_path)
    with open(vectors_fp, 'wb') as vectors_fh, open(metas_fp, 'wb') as metas_fh, open(
        ids_fp, 'w'
    ) as ids_fh:
        while shard_docs_written < size_this_shard:
            _write_shard_files(data, ids_fh, metas_fh, vectors_fh)
            shard_docs_written += 1


def _write_shard_files(
    data: Generator[Tuple[str, np.array, bytes], None, None],
    ids_fh: TextIO,
    metas_fh: BinaryIO,
    vectors_fh: BinaryIO,
):
    id_, vec, meta = next(data)
    vec_bytes = vec.tobytes()
    vectors_fh.write(len(vec_bytes).to_bytes(BYTE_PADDING, sys.byteorder) + vec_bytes)
    metas_fh.write(len(meta).to_bytes(BYTE_PADDING, sys.byteorder) + meta)
    ids_fh.write(id_ + '\n')


def import_vectors(path: str, pea_id: str):
    """Import id and vectors

    :param path: the path to the dump
    :param pea_id: the id of the pea (as part of the shards)
    :return: the generators for the ids and for the vectors
    """
    path = os.path.join(path, pea_id)
    ids_gen = _ids_gen(path)
    vecs_gen = _vecs_gen(path)
    return ids_gen, vecs_gen


def import_metas(path: str, pea_id: str):
    """Import id and metadata

    :param path: the path of the dump
    :param pea_id: the id of the pea (as part of the shards)
    :return: the generators for the ids and for the metadata
    """
    path = os.path.join(path, pea_id)
    ids_gen = _ids_gen(path)
    metas_gen = _metas_gen(path)
    return ids_gen, metas_gen


def _ids_gen(path: str):
    with open(os.path.join(path, 'ids'), 'r') as ids_fh:
        for l in ids_fh:
            yield l.strip()


def _vecs_gen(path: str):
    with open(os.path.join(path, 'vectors'), 'rb') as vectors_fh:
        while True:
            next_size = vectors_fh.read(BYTE_PADDING)
            next_size = int.from_bytes(next_size, byteorder=sys.byteorder)
            if next_size:
                vec = np.frombuffer(
                    vectors_fh.read(next_size),
                    dtype=np.float64,
                )
                yield vec
            else:
                break


def _metas_gen(path: str):
    with open(os.path.join(path, 'metas'), 'rb') as metas_fh:
        while True:
            next_size = metas_fh.read(BYTE_PADDING)
            next_size = int.from_bytes(next_size, byteorder=sys.byteorder)
            if next_size:
                meta = metas_fh.read(next_size)
                yield meta
            else:
                break


def _get_file_paths(shard_path: str):
    vectors_fp = os.path.join(shard_path, 'vectors')
    metas_fp = os.path.join(shard_path, 'metas')
    ids_fp = os.path.join(shard_path, 'ids')
    return vectors_fp, metas_fp, ids_fp



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import os
from typing import Optional

import numpy as np

from jina.executors.decorators import batching, as_ndarray
from jina.executors.encoders.frameworks import BaseTFEncoder
from jina.excepts import PretrainedModelFileDoesNotExist


class BigTransferEncoder(BaseTFEncoder):
    """
    :class:`BigTransferEncoder` is Big Transfer (BiT) presented by
    Google (https://github.com/google-research/big_transfer).
    Uses pretrained BiT to encode document content from a ndarray, potentially
    B x (Channel x Height x Width) into a ndarray of `B x D`.
    Internally, :class:`BigTransferEncoder` wraps the models from
    https://storage.googleapis.com/bit_models/.

    .. warning::

        Known issue: this does not work on tensorflow==2.2.0,
        https://github.com/tensorflow/tensorflow/issues/38571

    :param model_path: the path of the model in the `SavedModel` format.
        The pretrained model can be downloaded at
        wget https://storage.googleapis.com/bit_models/Imagenet21k/[model_name]/feature_vectors/saved_model.pb
        wget https://storage.googleapis.com/bit_models/Imagenet21k/[model_name]/feature_vectors/variables/variables.data-00000-of-00001
        wget https://storage.googleapis.com/bit_models/Imagenet21k/[model_name]/feature_vectors/variables/variables.index

        ``[model_name]`` includes `R50x1`, `R101x1`, `R50x3`, `R101x3`, `R152x4`

        The `model_path` should be a directory path, which has the following structure.

        .. highlight:: bash
         .. code-block:: bash

            .
            ├── saved_model.pb
            └── variables
                ├── variables.data-00000-of-00001
                └── variables.index

        :param channel_axis: the axis id of the channel, -1 indicate the color
            channel info at the last axis. If given other, then `
            `np.moveaxis(content, channel_axis, -1)`` is performed before :meth:`encode`.
    """

    def __init__(self,
                 model_path: Optional[str] = '/workspace/pretrained',
                 channel_axis: int = 1,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.channel_axis = channel_axis
        self.model_path = model_path

    def post_init(self):
        """Load model. Raise exception if model doesn't exist"""
        super().post_init()
        if self.model_path and os.path.exists(self.model_path):
            self.to_device()
            import tensorflow as tf
            self.logger.info(f'model_path: {self.model_path}')
            _model = tf.saved_model.load(self.model_path)
            self.model = _model.signatures['serving_default']
            self._get_input = tf.convert_to_tensor
        else:
            raise PretrainedModelFileDoesNotExist(f'model at {self.model_path} does not exist')

    @batching
    @as_ndarray
    def encode(self, content: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        """
        Encode content into a ndarray of `B x D`.
        Where `B` is the batch size and `D` is the Dimension.

        :param content: an array in size `B`
        :return: an ndarray in size `B x D`.
        """
        if self.channel_axis != -1:
            content = np.moveaxis(content, self.channel_axis, -1)
        _output = self.model(self._get_input(content.astype(np.float32)))
        return _output['output_1'].numpy()



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from collections import defaultdict
from typing import Tuple, Dict, List

import numpy as np

from . import FlatRecursiveMixin
from .encode import BaseEncodeDriver
from ..types.document.multimodal import MultimodalDocument

if False:
    from ..types.sets import DocumentSet


class MultiModalDriver(FlatRecursiveMixin, BaseEncodeDriver):
    """Extract multimodal embeddings from different modalities.

    Input-Output ::

        Input:
        document:
                |- chunk: {modality: mode1}
                |
                |- chunk: {modality: mode2}
        Output:
        document: (embedding: multimodal encoding)
                |- chunk: {modality: mode1}
                |
                |- chunk: {modality: mode2}

    .. note::

        - It traverses on the ``documents`` for which we want to apply the ``multimodal`` embedding. This way

        we can use the `batching` capabilities for the `executor`.

    .. warning::
        - It assumes that every ``chunk`` of a ``document`` belongs to a different modality.
    """

    def __init__(self, traversal_paths: Tuple[str] = ('r',), *args, **kwargs):
        super().__init__(traversal_paths=traversal_paths, *args, **kwargs)

    @property
    def positional_modality(self) -> List[str]:
        """Get position per modality.
        :return: the list of strings representing the name and order of the modality.
        """
        if not self._exec.positional_modality:
            raise RuntimeError(
                'Could not know which position of the ndarray to load to each modality'
            )
        return self._exec.positional_modality

    def _get_executor_input_arguments(
        self, content_by_modality: Dict[str, 'np.ndarray']
    ) -> List['np.ndarray']:
        """From a dictionary ``content_by_modality`` it returns the arguments in the proper order so that they can be
        passed to the executor.

         :param content_by_modality: a dictionary of `Document content` by modality name
         :return: list of input arguments as np arrays
        """
        return [content_by_modality[modality] for modality in self.positional_modality]

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs) -> None:
        """Apply the driver to each of the Documents in docs.

        :param docs: the docs for which a ``multimodal embedding`` will be computed, whose chunks are of different
        :param args: unused
        :param kwargs: unused
        """
        content_by_modality = defaultdict(
            list
        )  # array of num_rows equal to num_docs and num_columns equal to

        valid_docs = []
        for doc in docs:
            # convert to MultimodalDocument
            doc = MultimodalDocument(doc)
            if doc.modality_content_map:
                valid_docs.append(doc)
                for modality in self.positional_modality:
                    content_by_modality[modality].append(doc[modality])
            else:
                self.logger.warning(
                    f'Invalid doc {doc.id}. Only one chunk per modality is accepted'
                )

        if len(valid_docs) > 0:
            # Pass a variable length argument (one argument per array)
            for modality in self.positional_modality:
                content_by_modality[modality] = np.stack(content_by_modality[modality])

            # Guarantee that the arguments are provided to the executor in its desired order
            input_args = self._get_executor_input_arguments(content_by_modality)
            embeds = self.exec_fn(*input_args)
            if len(valid_docs) != embeds.shape[0]:
                self.logger.error(
                    f'mismatched {len(valid_docs)} docs from level {valid_docs[0].granularity} '
                    f'and a {embeds.shape} shape embedding, the first dimension must be the same'
                )
            for doc, embedding in zip(valid_docs, embeds):
                doc.embedding = embedding



import time
from subprocess import Popen, PIPE

from ..zmq.base import ZMQManyRuntime
from ....helper import ArgNamespace


class SSHRuntime(ZMQManyRuntime):
    """Simple SSH based SSHRuntime for remote Pea management
    .. note::
        It requires one to upload host public key to the remote
        1. ssh-keygen -b 4096
        2. scp ~/.ssh/id_rsa.pub username@hostname:~/.ssh/authorized_keys
    .. note::
        As the terminal signal is sent via :meth:`send_terminate_signal` from
        :class:`BasePea`, there is no need to override/implement :meth:`close`
        method. Lifecycle is handled by :class:`BasePea`.
    """

    def setup(self):
        """Setup the ssh communication to host."""
        self._ssh_proc = Popen(
            ['ssh', self.args.host],
            stdout=PIPE,
            stdin=PIPE,
            bufsize=0,
            universal_newlines=True,
        )
        self._ssh_proc.stdin.write(self._pea_command + '\n')
        while self._ssh_proc.poll() is None and not self.is_ready:
            time.sleep(1)

        # two cases to reach here: 1. is_ready, 2. container is dead
        if self._ssh_proc.poll() is not None:
            raise Exception(
                'the subprocess fails to start, check the arguments or entrypoint'
            )

    def run_forever(self):
        """Method to block the main thread and print logs."""
        for line in self._ssh_proc.stdout:
            self.logger.info(line.strip())

    def teardown(self):
        """Close the ssh communication."""
        self._ssh_proc.stdin.write('logout\n')
        self._ssh_proc.stdin.close()
        self._ssh_proc.stdout.close()
        super().teardown()

    @property
    def _pea_command(self) -> str:
        from jina.parsers import set_pea_parser

        non_defaults = ArgNamespace.get_non_defaults_args(
            self.args, set_pea_parser(), taboo={'host'}
        )
        _args = ArgNamespace.kwargs2list(non_defaults)
        return f'jina pea {" ".join(_args)}'



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Iterable, Tuple, Optional

from . import (
    BaseExecutableDriver,
    QuerySetReader,
    FlatRecursiveMixin,
    ContextAwareRecursiveMixin,
)
from ..types.document import Document
from ..types.score import NamedScore
from ..enums import EmbeddingClsType

if False:
    from ..types.sets import DocumentSet


class BaseSearchDriver(BaseExecutableDriver):
    """Drivers inherited from this Driver will bind :meth:`query` by default """

    def __init__(
        self,
        executor: Optional[str] = None,
        method: str = 'query',
        traversal_paths: Tuple[str] = ('r', 'c'),
        *args,
        **kwargs,
    ):
        super().__init__(
            executor, method, traversal_paths=traversal_paths, *args, **kwargs
        )


class KVSearchDriver(ContextAwareRecursiveMixin, BaseSearchDriver):
    """Fill in the results using the :class:`jina.executors.indexers.meta.BinaryPbIndexer`

    .. warning::
        This driver runs a query for each document.
        This may not be very efficient, as the total number of queries grows cubic with the number of documents, chunks
        per document and top-k.

            - traversal_paths = ['m'] => D x K
            - traversal_paths = ['r'] => D
            - traversal_paths = ['cm'] => D x C x K
            - traversal_paths = ['m', 'cm'] => D x K + D x C x K

        where:
            - D is the number of queries
            - C is the number of chunks per document
            - K is the top-k

        :param is_update: when set to true the retrieved docs are merged into current message;
            otherwise, the retrieved Document overrides the existing Document
        :param traversal_paths: traversal paths for the driver
        :param args: additional positional arguments which are just used for the parent initialization
        :param kwargs: additional key value arguments which are just used for the parent initialization
    """

    def __init__(
        self,
        is_update: bool = True,
        traversal_paths: Tuple[str] = ('m',),
        *args,
        **kwargs,
    ):
        super().__init__(traversal_paths=traversal_paths, *args, **kwargs)
        self._is_update = is_update

    def _apply_all(
        self, doc_sequences: Iterable['DocumentSet'], *args, **kwargs
    ) -> None:
        for docs in doc_sequences:
            miss_idx = (
                []
            )  #: missed hit results, some search may not end with results. especially in shards
            for idx, retrieved_doc in enumerate(docs):
                serialized_doc = self.exec_fn(retrieved_doc.id)
                if serialized_doc:
                    r = Document(serialized_doc)
                    if self._is_update:
                        retrieved_doc.update(r)
                    else:
                        retrieved_doc.CopyFrom(r)
                else:
                    miss_idx.append(idx)
            # delete non-existed matches in reverse
            for j in reversed(miss_idx):
                del docs[j]


class VectorFillDriver(FlatRecursiveMixin, QuerySetReader, BaseSearchDriver):
    """Fill in the embedding by their document id."""

    def __init__(
        self,
        executor: Optional[str] = None,
        method: str = 'query_by_key',
        *args,
        **kwargs,
    ):
        super().__init__(executor, method, *args, **kwargs)

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs) -> None:
        embeds = self.exec_fn([d.id for d in docs])
        for doc, embedding in zip(docs, embeds):
            doc.embedding = embedding


class VectorSearchDriver(FlatRecursiveMixin, QuerySetReader, BaseSearchDriver):
    """Extract dense embeddings from the request for the executor to query.

    :param top_k: top-k document ids to retrieve
    :param fill_embedding: fill in the embedding of the corresponding doc,
            this requires the executor to implement :meth:`query_by_key`
    :param args: additional positional arguments which are just used for the parent initialization
    :param kwargs: additional key value arguments which are just used for the parent initialization"""

    def __init__(self, top_k: int = 50, fill_embedding: bool = False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._top_k = top_k
        self._fill_embedding = fill_embedding

    @property
    def exec_embedding_cls_type(self) -> EmbeddingClsType:
        """Get the sparse class type of the attached executor.

        :return: Embedding class type of the attached executor, default value is `dense`
        """
        return EmbeddingClsType.from_string(self.exec.embedding_cls_type)

    def _get_documents_embeddings(self, docs: 'DocumentSet'):
        embedding_cls_type = self.exec_embedding_cls_type
        if embedding_cls_type.is_dense:
            return docs.all_embeddings
        else:
            return docs.get_all_sparse_embeddings(embedding_cls_type=embedding_cls_type)

    def _fill_matches(self, doc, op_name, topks, scores, topk_embed):
        embedding_cls_type = self.exec_embedding_cls_type
        if embedding_cls_type.is_dense:
            for numpy_match_id, score, vector in zip(topks, scores, topk_embed):
                m = Document(id=numpy_match_id)
                m.score = NamedScore(op_name=op_name, value=score)
                r = doc.matches.append(m)
                if vector is not None:
                    r.embedding = vector
        else:
            for idx, (numpy_match_id, score) in enumerate(zip(topks, scores)):
                vector = None
                if topk_embed[idx] is not None:
                    vector = topk_embed.getrow(idx)
                m = Document(id=numpy_match_id)
                m.score = NamedScore(op_name=op_name, value=score)
                match = doc.matches.append(m)
                if vector is not None:
                    match.embedding = vector

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs) -> None:
        embed_vecs, doc_pts = self._get_documents_embeddings(docs)

        if not doc_pts:
            return

        fill_fn = getattr(self.exec, 'query_by_key', None)
        if self._fill_embedding and not fill_fn:
            self.logger.warning(
                f'"fill_embedding=True" but {self.exec} does not have "query_by_key" method'
            )

        idx, dist = self.exec_fn(embed_vecs, top_k=int(self.top_k))
        op_name = self.exec.__class__.__name__
        for doc, topks, scores in zip(doc_pts, idx, dist):
            topk_embed = (
                fill_fn(topks)
                if (self._fill_embedding and fill_fn)
                else [None] * len(topks)
            )
            self._fill_matches(doc, op_name, topks, scores, topk_embed)



from typing import Optional

from . import DocsExtractUpdateMixin, FlatRecursiveMixin, BaseExecutableDriver


class GenericExecutorDriver(
    DocsExtractUpdateMixin, FlatRecursiveMixin, BaseExecutableDriver
):
    """Generic driver that uses extract-apply-update pattern. It automatically binds to the method
    decorated with `@request`."""

    def __init__(
        self, executor: Optional[str] = None, method: str = '', *args, **kwargs
    ):
        super().__init__(executor, method, *args, **kwargs)



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import argparse
from argparse import Namespace
from contextlib import ExitStack
from typing import Optional, Dict, List, Union, Set

from jina.peapods.zmq import send_ctrl_message
from jina.types.message.dump import DumpMessage

import copy
from argparse import Namespace
from typing import List, Optional
from itertools import cycle

from ... import __default_host__
from ...enums import SchedulerType, SocketType, PeaRoleType, PollingType
from ...helper import get_public_ip, get_internal_ip, random_identity
from ... import helper

from ..peas import BasePea
from ...enums import SchedulerType, PodRoleType


class BasePod(ExitStack):
    """A BasePod is an immutable set of peas. They share the same input and output socket.
    Internally, the peas can run with the process/thread backend.
    They can be also run in their own containers on remote machines.
    """

    def __init__(self, args: Union['argparse.Namespace', Dict], needs: Set[str] = None):
        super().__init__()
        self.peas = []  # type: List['BasePea']
        self.args = args
        self._set_conditional_args(self.args)
        self.needs = (
            needs if needs else set()
        )  #: used in the :class:`jina.flow.Flow` to build the graph

        self.is_head_router = False
        self.is_tail_router = False
        self.deducted_head = None
        self.deducted_tail = None

    def start(self) -> 'BasePod':
        """Start to run all :class:`BasePea` in this BasePod.

        .. note::
            If one of the :class:`BasePea` fails to start, make sure that all of them
            are properly closed.
        """
        raise NotImplemented()

    def close(self):
        """Stop all :class:`BasePea` in this BasePod."""
        self.__exit__(None, None, None)

    @staticmethod
    def _set_conditional_args(args):
        if args.pod_role == PodRoleType.GATEWAY:
            if args.restful:
                args.runtime_cls = 'RESTRuntime'
            else:
                args.runtime_cls = 'GRPCRuntime'

    @property
    def role(self) -> 'PodRoleType':
        """Return the role of this :class:`BasePod`.

        .. # noqa: DAR201
        """
        return self.args.pod_role

    @property
    def name(self) -> str:
        """The name of this :class:`BasePod`.


        .. # noqa: DAR201
        """
        return self.args.name

    @property
    def host_in(self) -> str:
        """Get the host_in of this pod


        .. # noqa: DAR201
        """
        return self.head_args.host_in

    @property
    def host_out(self) -> str:
        """Get the host_out of this pod


        .. # noqa: DAR201
        """
        return self.tail_args.host_out

    @property
    def address_in(self) -> str:
        """Get the full incoming address of this pod


        .. # noqa: DAR201
        """
        return f'{self.head_args.host_in}:{self.head_args.port_in} ({self.head_args.socket_in!s})'

    @property
    def address_out(self) -> str:
        """Get the full outgoing address of this pod


        .. # noqa: DAR201
        """
        return f'{self.tail_args.host_out}:{self.tail_args.port_out} ({self.tail_args.socket_out!s})'

    def _parse_base_pod_args(
        self,
        args,
        attribute,
        id_attribute_name,
        role_type,
        repetition_attribute,
        polling_type,
    ):
        parsed_args = {'head': None, 'tail': None, attribute: []}
        if getattr(args, repetition_attribute, 1) > 1:
            # reasons to separate head and tail from peas is that they
            # can be deducted based on the previous and next pods
            self._set_after_to_pass(args)
            self.is_head_router = True
            self.is_tail_router = True
            parsed_args['head'] = BasePod._copy_to_head_args(args, polling_type.is_push)
            parsed_args['tail'] = BasePod._copy_to_tail_args(args, polling_type.is_push)
            parsed_args[attribute] = BasePod._set_peas_args(
                args,
                role_type=role_type,
                repetition_attribute=repetition_attribute,
                id_attribute_name=id_attribute_name,
                polling_type=polling_type,
                head_args=parsed_args['head'],
                tail_args=parsed_args['tail'],
            )
        elif (getattr(args, 'uses_before', None) and args.uses_before != '_pass') or (
            getattr(args, 'uses_after', None) and args.uses_after != '_pass'
        ):
            args.scheduling = SchedulerType.ROUND_ROBIN
            if getattr(args, 'uses_before', None):
                self.is_head_router = True
                parsed_args['head'] = self._copy_to_head_args(
                    args, polling_type.is_push
                )
            if getattr(args, 'uses_after', None):
                self.is_tail_router = True
                parsed_args['tail'] = self._copy_to_tail_args(
                    args, polling_type.is_push
                )
            parsed_args[attribute] = self._set_peas_args(
                args,
                role_type=role_type,
                repetition_attribute=repetition_attribute,
                id_attribute_name=id_attribute_name,
                polling_type=polling_type,
                head_args=parsed_args.get('head', None),
                tail_args=parsed_args.get('tail', None),
            )
        else:
            self.is_head_router = False
            self.is_tail_router = False
            parsed_args[attribute] = [args]

        # note that peas_args['peas'][0] exist either way and carries the original property
        return parsed_args

    def _enter_pea(self, pea: 'BasePea') -> None:
        self.peas.append(pea)
        self.enter_context(pea)

    def __enter__(self) -> 'BasePod':
        return self.start()

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        super().__exit__(exc_type, exc_val, exc_tb)
        self.join()

    @staticmethod
    def _set_peas_args(
        args: Namespace,
        role_type: PeaRoleType,
        repetition_attribute: str,
        id_attribute_name: str,
        polling_type: PollingType,
        head_args: Optional[Namespace] = None,
        tail_args: Namespace = None,
    ) -> List[Namespace]:
        result = []
        _host_list = (
            args.peas_hosts
            if args.peas_hosts
            else [
                args.host,
            ]
        )

        for idx, pea_host in zip(
            range(getattr(args, repetition_attribute)), cycle(_host_list)
        ):
            _args = copy.deepcopy(args)

            setattr(_args, id_attribute_name, idx)

            if getattr(args, repetition_attribute) > 1:
                _args.pea_role = role_type
                _args.identity = random_identity()
                if _args.peas_hosts:
                    _args.host = pea_host
                if _args.name:
                    _args.name += f'/{idx}'
                else:
                    _args.name = f'{idx}'
            else:
                _args.pea_role = PeaRoleType.SINGLETON

            if head_args:
                _args.port_in = head_args.port_out
            if tail_args:
                _args.port_out = tail_args.port_in
            _args.port_ctrl = helper.random_port()
            _args.socket_out = SocketType.PUSH_CONNECT
            if polling_type.is_push:
                if args.scheduling == SchedulerType.ROUND_ROBIN:
                    _args.socket_in = SocketType.PULL_CONNECT
                elif args.scheduling == SchedulerType.LOAD_BALANCE:
                    _args.socket_in = SocketType.DEALER_CONNECT
                else:
                    raise ValueError(
                        f'{args.scheduling} is not supported as a SchedulerType!'
                    )

            else:
                _args.socket_in = SocketType.SUB_CONNECT
            if head_args:
                _args.host_in = BasePod._fill_in_host(
                    bind_args=head_args, connect_args=_args
                )
            if tail_args:
                _args.host_out = BasePod._fill_in_host(
                    bind_args=tail_args, connect_args=_args
                )

            result.append(_args)
        return result

    @staticmethod
    def _set_after_to_pass(args):
        raise NotImplemented

    @staticmethod
    def _copy_to_head_args(
        args: Namespace, is_push: bool, as_router: bool = True
    ) -> Namespace:
        """
        Set the outgoing args of the head router

        :param args: basic arguments
        :param is_push: if true, set socket_out based on the SchedulerType
        :param as_router: if true, router configuration is applied
        :return: enriched head arguments
        """

        _head_args = copy.deepcopy(args)
        _head_args.port_ctrl = helper.random_port()
        _head_args.port_out = helper.random_port()
        _head_args.uses = None
        if is_push:
            if args.scheduling == SchedulerType.ROUND_ROBIN:
                _head_args.socket_out = SocketType.PUSH_BIND
            elif args.scheduling == SchedulerType.LOAD_BALANCE:
                _head_args.socket_out = SocketType.ROUTER_BIND
        else:
            _head_args.socket_out = SocketType.PUB_BIND
        if as_router:
            _head_args.uses = args.uses_before or '_pass'

        if as_router:
            _head_args.pea_role = PeaRoleType.HEAD
            if args.name:
                _head_args.name = f'{args.name}/head'
            else:
                _head_args.name = f'head'

        # in any case, if header is present, it represent this Pod to consume `num_part`
        # the following peas inside the pod will have num_part=1
        args.num_part = 1

        return _head_args

    @staticmethod
    def _copy_to_tail_args(
        args: Namespace, is_push: bool, as_router: bool = True
    ) -> Namespace:
        """
        Set the incoming args of the tail router

        :param args: configuration for the connection
        :param is_push: if true, set socket_out based on the SchedulerType
        :param as_router: if true, add router configuration
        :return: enriched arguments
        """
        _tail_args = copy.deepcopy(args)
        _tail_args.port_in = helper.random_port()
        _tail_args.port_ctrl = helper.random_port()
        _tail_args.socket_in = SocketType.PULL_BIND
        _tail_args.uses = None

        if as_router:
            _tail_args.uses = args.uses_after or '_pass'
            if args.name:
                _tail_args.name = f'{args.name}/tail'
            else:
                _tail_args.name = f'tail'
            _tail_args.pea_role = PeaRoleType.TAIL
            _tail_args.num_part = 1 if is_push else args.parallel

        return _tail_args

    @staticmethod
    def _fill_in_host(bind_args: Namespace, connect_args: Namespace) -> str:
        """
        Compute the host address for ``connect_args``

        :param bind_args: configuration for the host ip binding
        :param connect_args: configuration for the host ip connection
        :return: host ip
        """
        from sys import platform

        # by default __default_host__ is 0.0.0.0

        # is BIND at local
        bind_local = bind_args.host == __default_host__

        # is CONNECT at local
        conn_local = connect_args.host == __default_host__

        # is CONNECT inside docker?
        conn_docker = getattr(
            connect_args, 'uses', None
        ) is not None and connect_args.uses.startswith('docker://')

        # is BIND & CONNECT all on the same remote?
        bind_conn_same_remote = (
            not bind_local and not conn_local and (bind_args.host == connect_args.host)
        )

        if platform in ('linux', 'linux2'):
            local_host = __default_host__
        else:
            local_host = 'host.docker.internal'

        # pod1 in local, pod2 in local (conn_docker if pod2 in docker)
        if bind_local and conn_local:
            return local_host if conn_docker else __default_host__

        # pod1 and pod2 are remote but they are in the same host (pod2 is local w.r.t pod1)
        if bind_conn_same_remote:
            return local_host if conn_docker else __default_host__

        # From here: Missing consideration of docker
        if bind_local and not conn_local:
            # in this case we are telling CONN (at remote) our local ip address
            return get_public_ip() if bind_args.expose_public else get_internal_ip()
        else:
            # in this case we (at local) need to know about remote the BIND address
            return bind_args.host


class Pod(BasePod):
    """A BasePod is an immutable set of peas, which run in parallel. They share the same input and output socket.
    Internally, the peas can run with the process/thread backend. They can be also run in their own containers
    :param args: arguments parsed from the CLI
    :param needs: pod names of preceding pods, the output of these pods are going into the input of this pod
    """

    def __init__(self, args: Union['argparse.Namespace', Dict], needs: Set[str] = None):
        super().__init__(args, needs)
        if isinstance(args, Dict):
            # This is used when a Pod is created in a remote context, where peas & their connections are already given.
            self.peas_args = args
        else:
            self.peas_args = self._parse_args(args)

    @property
    def is_singleton(self) -> bool:
        """Return if the Pod contains only a single Pea


        .. # noqa: DAR201
        """
        return not (self.is_head_router or self.is_tail_router)

    @property
    def first_pea_args(self) -> Namespace:
        """Return the first non-head/tail pea's args


        .. # noqa: DAR201
        """
        # note this will be never out of boundary
        return self.peas_args['peas'][0]

    @property
    def port_expose(self) -> int:
        """Get the grpc port number


        .. # noqa: DAR201
        """
        return self.first_pea_args.port_expose

    @property
    def host(self) -> str:
        """Get the host name of this Pod


        .. # noqa: DAR201
        """
        return self.first_pea_args.host

    def _parse_args(
        self, args: Namespace
    ) -> Dict[str, Optional[Union[List[Namespace], Namespace]]]:
        return self._parse_base_pod_args(
            args,
            attribute='peas',
            id_attribute_name='pea_id',
            role_type=PeaRoleType.PARALLEL,
            repetition_attribute='parallel',
            polling_type=getattr(args, 'polling', None),
        )

    @property
    def head_args(self):
        """Get the arguments for the `head` of this BasePod.


        .. # noqa: DAR201
        """
        if self.is_head_router and self.peas_args['head']:
            return self.peas_args['head']
        elif not self.is_head_router and len(self.peas_args['peas']) == 1:
            return self.first_pea_args
        elif self.deducted_head:
            return self.deducted_head
        else:
            raise ValueError('ambiguous head node, maybe it is deducted already?')

    @head_args.setter
    def head_args(self, args):
        """Set the arguments for the `head` of this BasePod.


        .. # noqa: DAR101
        """
        if self.is_head_router and self.peas_args['head']:
            self.peas_args['head'] = args
        elif not self.is_head_router and len(self.peas_args['peas']) == 1:
            self.peas_args['peas'][0] = args  # weak reference
        elif self.deducted_head:
            self.deducted_head = args
        else:
            raise ValueError('ambiguous head node, maybe it is deducted already?')

    @property
    def tail_args(self):
        """Get the arguments for the `tail` of this BasePod.

        .. # noqa: DAR201
        """
        if self.is_tail_router and self.peas_args['tail']:
            return self.peas_args['tail']
        elif not self.is_tail_router and len(self.peas_args['peas']) == 1:
            return self.first_pea_args
        elif self.deducted_tail:
            return self.deducted_tail
        else:
            raise ValueError('ambiguous tail node, maybe it is deducted already?')

    @tail_args.setter
    def tail_args(self, args):
        """Set the arguments for the `tail` of this BasePod.

        .. # noqa: DAR101
        """
        if self.is_tail_router and self.peas_args['tail']:
            self.peas_args['tail'] = args
        elif not self.is_tail_router and len(self.peas_args['peas']) == 1:
            self.peas_args['peas'][0] = args  # weak reference
        elif self.deducted_tail:
            self.deducted_tail = args
        else:
            raise ValueError('ambiguous tail node, maybe it is deducted already?')

    @property
    def all_args(self) -> List[Namespace]:
        """Get all arguments of all Peas in this BasePod.

        .. # noqa: DAR201
        """
        return (
            ([self.peas_args['head']] if self.peas_args['head'] else [])
            + ([self.peas_args['tail']] if self.peas_args['tail'] else [])
            + self.peas_args['peas']
        )

    @property
    def num_peas(self) -> int:
        """Get the number of running :class:`BasePea`

        .. # noqa: DAR201
        """
        return len(self.peas)

    def __eq__(self, other: 'BasePod'):
        return self.num_peas == other.num_peas and self.name == other.name

    def start(self) -> 'BasePod':
        """
        Start to run all :class:`BasePea` in this BasePod.

        :return: started pod

        .. note::
            If one of the :class:`BasePea` fails to start, make sure that all of them
            are properly closed.
        """
        if getattr(self.args, 'noblock_on_start', False):
            for _args in self.all_args:
                _args.noblock_on_start = True
                self._enter_pea(BasePea(_args))
            # now rely on higher level to call `wait_start_success`
            return self
        else:
            try:
                for _args in self.all_args:
                    self._enter_pea(BasePea(_args))
            except:
                self.close()
                raise
            return self

    def wait_start_success(self) -> None:
        """Block until all peas starts successfully.

        If not successful, it will raise an error hoping the outer function to catch it
        """

        if not self.args.noblock_on_start:
            raise ValueError(
                f'{self.wait_start_success!r} should only be called when `noblock_on_start` is set to True'
            )

        try:
            for p in self.peas:
                p.wait_start_success()
        except:
            self.close()
            raise

    def join(self):
        """Wait until all peas exit"""
        try:
            for p in self.peas:
                p.join()
        except KeyboardInterrupt:
            pass
        finally:
            self.peas.clear()

    @property
    def is_ready(self) -> bool:
        """Checks if Pod is ready

        .. note::
            A Pod is ready when all the Peas it contains are ready


        .. # noqa: DAR201
        """
        return all(p.is_ready.is_set() for p in self.peas)

    def _set_after_to_pass(self, args):
        # TODO: check if needed
        # remark 1: i think it's related to route driver.
        if hasattr(args, 'polling') and args.polling.is_push:
            # ONLY reset when it is push
            args.uses_after = '_pass'

    def dump(self, path, shards, timeout):
        """Emit a Dump request to its Peas

        :param shards: the nr of shards in the dump
        :param path: the path to which to dump
        :param timeout: time to wait (seconds)
        """
        for pea in self.peas:
            if 'head' not in pea.name and 'tail' not in pea.name:
                send_ctrl_message(
                    pea.runtime.ctrl_addr,
                    DumpMessage(path=path, shards=shards),
                    timeout=timeout,
                )



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from .document import DocumentSet
from .querylang import QueryLangSet
from .chunk import ChunkSet
from .match import MatchSet



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Tuple, Generator
import numpy as np

from jina.executors.indexers.dbms import BaseDBMSIndexer
from jina.executors.indexers.dump import export_dump_streaming

from .postgreshandler import PostgreSQLDBMSHandler


class PostgreSQLDBMSIndexer(BaseDBMSIndexer):
    """:class:`PostgreSQLDBMSIndexer` PostgreSQL based BDMS Indexer.
    Initialize the PostgreSQLDBIndexer.

    :param hostname: hostname of the machine
    :param port: the port
    :param username: the username to authenticate
    :param password: the password to authenticate
    :param database: the database name
    :param table: the table name to use
    :param args: other arguments
    :param kwargs: other keyword arguments
    """

    def __init__(
            self,
            hostname: str = '127.0.0.1',
            port: int = 5432,
            username: str = 'postgres',
            password: str = '123456',
            database: str = 'postgres',
            table: str = 'default_table',
            *args,
            **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.hostname = hostname
        self.port = port
        self.username = username
        self.password = password
        self.database = database
        self.table = table

    def _get_generator(self) -> Generator[Tuple[str, np.array, bytes], None, None]:
        with self.handler as handler:
            # always order the dump by id as integer
            handler.cursor.execute(f"SELECT * from {handler.table} ORDER BY ID::int")
            records = handler.cursor.fetchall()
            for rec in records:
                yield rec[0], rec[1], rec[2]

    @property
    def size(self):
        """Obtain the size of the table

        .. # noqa: DAR201
        """
        with self.handler as postgres_handler:
            postgres_handler.cursor.execute(f"SELECT COUNT(*) from {self.handler.table}")
            records = postgres_handler.cursor.fetchall()
            return records[0][0]

    def post_init(self):
        """Initialize the PostgresHandler inside the Indexer."""
        from .postgreshandler import PostgreSQLDBMSHandler
        super().post_init()
        self.handler = PostgreSQLDBMSHandler(
            hostname=self.hostname,
            port=self.port,
            username=self.username,
            password=self.password,
            database=self.database,
            table=self.table)

    def get_handler(self) -> 'PostgreSQLDBMSHandler':
        """Get the handler to PostgreSQLDBMS."""
        return self.handler

    def get_add_handler(self) -> 'PostgreSQLDBMSHandler':
        """Get the handler to PostgresSQLDBMS."""
        return self.handler

    def get_create_handler(self) -> 'PostgreSQLDBMSHandler':
        """Get the handler to PostgresSQLDBMS."""
        return self.handler

    def get_query_handler(self) -> 'PostgreSQLDBMSHandler':
        """Get the handler to PostgresSQLDBMS."""
        return self.handler

    def add(self, ids, vecs, metas, *args, **kwargs):
        """Add a Document to PostgreSQLDBMS.

        :param ids: List of doc ids to be added
        :param vecs: List of vecs to be added
        :param metas: List of metas of docs to be added
         """
        with self.handler as postgres_handler:
            postgres_handler.add(ids=ids, vecs=vecs, metas=metas)

    def update(self, ids, vecs, metas, *args, **kwargs):
        """Updated document from the database.

        :param ids: Ids of Docs to be updated
        :param vecs: List of vecs to be updated
        :param metas: List of metas of docs to be updated
        """

        with self.handler as postgres_handler:
            postgres_handler.update(ids=ids, vecs=vecs, metas=metas)

    def delete(self, ids, *args, **kwargs):
        """Delete document from the database.

        :param ids: Ids of Document to be removed
        """

        with self.handler as postgres_handler:
            postgres_handler.delete(ids=ids)

    def dump(self, path, shards):
        """Dump the index

        :param path: the path to which to dump
        :param shards: the nr of shards to which to dump
        """
        export_dump_streaming(
            path,
            shards=shards,
            size=self.size,
            data=self._get_generator()
        )



from ..importer import IMPORTED

schema_all_drivers = {
    'Jina::Drivers::All': {
        'type': 'array',
        'items': {
            'oneOf': [
                {'$ref': f'#/definitions/{k}'} for k in IMPORTED.schema_drivers.keys()
            ]
        },
        'minItems': 1,
    }
}



import warnings
from typing import Union, Iterable, TextIO, Dict, Optional

import numpy as np

from ...clients.base import InputType, CallbackFnType
from ...enums import DataInputType
from ...helper import deprecated_alias


class AsyncCRUDFlowMixin:
    """The asynchronous version of the Mixin for CRUD in Flow"""

    @deprecated_alias(
        input_fn=('inputs', 0),
        buffer=('inputs', 1),
        callback=('on_done', 1),
        output_fn=('on_done', 1),
    )
    async def train(
        self,
        inputs: InputType,
        on_done: CallbackFnType = None,
        on_error: CallbackFnType = None,
        on_always: CallbackFnType = None,
        **kwargs,
    ):
        """Do training on the current Flow

        :param inputs: An iterator of bytes. If not given, then you have to specify it in **kwargs**.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :yields: results
        """
        warnings.warn(f'{self.train} is under heavy refactoring', FutureWarning)
        async for r in self._get_client(**kwargs).train(
            inputs, on_done, on_error, on_always, **kwargs
        ):
            yield r

    @deprecated_alias(
        input_fn=('inputs', 0),
        buffer=('inputs', 1),
        callback=('on_done', 1),
        output_fn=('on_done', 1),
    )
    async def index_ndarray(
        self,
        array: 'np.ndarray',
        axis: int = 0,
        size: Optional[int] = None,
        shuffle: bool = False,
        on_done: CallbackFnType = None,
        on_error: CallbackFnType = None,
        on_always: CallbackFnType = None,
        **kwargs,
    ):
        """Using numpy ndarray as the index source for the current Flow

        :param array: the numpy ndarray data source
        :param axis: iterate over that axis
        :param size: the maximum number of the sub arrays
        :param shuffle: shuffle the the numpy data source beforehand
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :yields: results
        """
        from ...clients.sugary_io import _input_ndarray

        async for r in self._get_client(**kwargs).index(
            _input_ndarray(array, axis, size, shuffle),
            on_done,
            on_error,
            on_always,
            data_type=DataInputType.CONTENT,
            **kwargs,
        ):
            yield r

    @deprecated_alias(
        input_fn=('inputs', 0),
        buffer=('inputs', 1),
        callback=('on_done', 1),
        output_fn=('on_done', 1),
    )
    async def search_ndarray(
        self,
        array: 'np.ndarray',
        axis: int = 0,
        size: Optional[int] = None,
        shuffle: bool = False,
        on_done: CallbackFnType = None,
        on_error: CallbackFnType = None,
        on_always: CallbackFnType = None,
        **kwargs,
    ):
        """Use a numpy ndarray as the query source for searching on the current Flow

        :param array: the numpy ndarray data source
        :param axis: iterate over that axis
        :param size: the maximum number of the sub arrays
        :param shuffle: shuffle the the numpy data source beforehand
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :yields: results
        """
        from ...clients.sugary_io import _input_ndarray

        async for r in self._get_client(**kwargs).search(
            _input_ndarray(array, axis, size, shuffle),
            on_done,
            on_error,
            on_always,
            data_type=DataInputType.CONTENT,
            **kwargs,
        ):
            yield r

    @deprecated_alias(
        input_fn=('inputs', 0),
        buffer=('inputs', 1),
        callback=('on_done', 1),
        output_fn=('on_done', 1),
    )
    async def index_lines(
        self,
        lines: Optional[Union[Iterable[str], TextIO]] = None,
        filepath: Optional[str] = None,
        size: Optional[int] = None,
        sampling_rate: Optional[float] = None,
        read_mode: str = 'r',
        line_format: str = 'json',
        field_resolver: Optional[Dict[str, str]] = None,
        on_done: CallbackFnType = None,
        on_error: CallbackFnType = None,
        on_always: CallbackFnType = None,
        **kwargs,
    ):
        """Use a list of lines as the index source for indexing on the current Flow

        :param lines: a list of strings, each is considered as d document
        :param filepath: a text file that each line contains a document
        :param size: the maximum number of the documents
        :param sampling_rate: the sampling rate between [0, 1]
        :param read_mode: specifies the mode in which the file
                is opened. 'r' for reading in text mode, 'rb' for reading in binary
        :param line_format: the format of each line: ``json`` or ``csv``
        :param field_resolver: a map from field names defined in ``document`` (JSON, dict) to the field
            names defined in Protobuf. This is only used when the given ``document`` is
            a JSON string or a Python dict.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :yields: results
        """
        from ...clients.sugary_io import _input_lines

        async for r in self._get_client(**kwargs).index(
            _input_lines(
                lines,
                filepath,
                size=size,
                sampling_rate=sampling_rate,
                read_mode=read_mode,
                line_format=line_format,
                field_resolver=field_resolver,
            ),
            on_done,
            on_error,
            on_always,
            data_type=DataInputType.AUTO,
            **kwargs,
        ):
            yield r

    async def index_csv(
        self,
        lines: Union[Iterable[str], TextIO],
        field_resolver: Dict[str, str] = None,
        size: Optional[int] = None,
        sampling_rate: Optional[float] = None,
        on_done: CallbackFnType = None,
        on_error: CallbackFnType = None,
        on_always: CallbackFnType = None,
        **kwargs,
    ):
        """Use a list of lines as the index source for indexing on the current Flow
        :param lines: a list of strings, each is considered as d document
        :param size: the maximum number of the documents
        :param sampling_rate: the sampling rate between [0, 1]
        :param field_resolver: a map from field names defined in ``document`` (JSON, dict) to the field
            names defined in Protobuf. This is only used when the given ``document`` is
            a JSON string or a Python dict.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :yields: results
        """
        from ...clients.sugary_io import _input_csv

        async for r in self._get_client(**kwargs).index(
            _input_csv(
                lines,
                size=size,
                sampling_rate=sampling_rate,
                field_resolver=field_resolver,
            ),
            on_done,
            on_error,
            on_always,
            data_type=DataInputType.AUTO,
            **kwargs,
        ):
            yield r

    async def index_ndjson(
        self,
        lines: Union[Iterable[str], TextIO],
        field_resolver: Optional[Dict[str, str]] = None,
        size: Optional[int] = None,
        sampling_rate: Optional[float] = None,
        on_done: CallbackFnType = None,
        on_error: CallbackFnType = None,
        on_always: CallbackFnType = None,
        **kwargs,
    ):
        """Use a list of lines as the index source for indexing on the current Flow
        :param lines: a list of strings, each is considered as d document
        :param size: the maximum number of the documents
        :param sampling_rate: the sampling rate between [0, 1]
        :param field_resolver: a map from field names defined in ``document`` (JSON, dict) to the field
            names defined in Protobuf. This is only used when the given ``document`` is
            a JSON string or a Python dict.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :yields: results
        """
        from ...clients.sugary_io import _input_ndjson

        async for r in self._get_client(**kwargs).index(
            _input_ndjson(
                lines,
                size=size,
                sampling_rate=sampling_rate,
                field_resolver=field_resolver,
            ),
            on_done,
            on_error,
            on_always,
            data_type=DataInputType.AUTO,
            **kwargs,
        ):
            yield r

    @deprecated_alias(
        input_fn=('inputs', 0),
        buffer=('inputs', 1),
        callback=('on_done', 1),
        output_fn=('on_done', 1),
    )
    async def index_files(
        self,
        patterns: Union[str, Iterable[str]],
        recursive: bool = True,
        size: Optional[int] = None,
        sampling_rate: Optional[float] = None,
        read_mode: Optional[str] = None,
        on_done: CallbackFnType = None,
        on_error: CallbackFnType = None,
        on_always: CallbackFnType = None,
        **kwargs,
    ):
        """Use a set of files as the index source for indexing on the current Flow

        :param patterns: The pattern may contain simple shell-style wildcards, e.g. '\*.py', '[\*.zip, \*.gz]'
        :param recursive: If recursive is true, the pattern '**' will match any files and
                    zero or more directories and subdirectories.
        :param size: the maximum number of the files
        :param sampling_rate: the sampling rate between [0, 1]
        :param read_mode: specifies the mode in which the file
                is opened. 'r' for reading in text mode, 'rb' for reading in binary mode
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :yields: results
        """
        from ...clients.sugary_io import _input_files

        async for r in self._get_client(**kwargs).index(
            _input_files(patterns, recursive, size, sampling_rate, read_mode),
            on_done,
            on_error,
            on_always,
            data_type=DataInputType.CONTENT,
            **kwargs,
        ):
            yield r

    @deprecated_alias(
        input_fn=('inputs', 0),
        buffer=('inputs', 1),
        callback=('on_done', 1),
        output_fn=('on_done', 1),
    )
    async def search_files(
        self,
        patterns: Union[str, Iterable[str]],
        recursive: bool = True,
        size: Optional[int] = None,
        sampling_rate: Optional[float] = None,
        read_mode: Optional[str] = None,
        on_done: CallbackFnType = None,
        on_error: CallbackFnType = None,
        on_always: CallbackFnType = None,
        **kwargs,
    ):
        """Use a set of files as the query source for searching on the current Flow

        :param patterns: The pattern may contain simple shell-style wildcards, e.g. '\*.py', '[\*.zip, \*.gz]'
        :param recursive: If recursive is true, the pattern '**' will match any files and
                    zero or more directories and subdirectories.
        :param size: the maximum number of the files
        :param sampling_rate: the sampling rate between [0, 1]
        :param read_mode: specifies the mode in which the file
                is opened. 'r' for reading in text mode, 'rb' for reading in
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :yields: results
        """
        from ...clients.sugary_io import _input_files

        async for r in self._get_client(**kwargs).search(
            _input_files(patterns, recursive, size, sampling_rate, read_mode),
            on_done,
            on_error,
            on_always,
            data_type=DataInputType.CONTENT,
            **kwargs,
        ):
            yield r

    async def search_ndjson(
        self,
        lines: Union[Iterable[str], TextIO],
        field_resolver: Optional[Dict[str, str]] = None,
        size: Optional[int] = None,
        sampling_rate: Optional[float] = None,
        on_done: CallbackFnType = None,
        on_error: CallbackFnType = None,
        on_always: CallbackFnType = None,
        **kwargs,
    ):
        """Use a list of files as the query source for searching on the current Flow
        :param lines: a list of strings, each is considered as d document
        :param size: the maximum number of the documents
        :param sampling_rate: the sampling rate between [0, 1]
        :param field_resolver: a map from field names defined in ``document`` (JSON, dict) to the field
            names defined in Protobuf. This is only used when the given ``document`` is
            a JSON string or a Python dict.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :yields: results
        """
        from ...clients.sugary_io import _input_ndjson

        async for r in self._get_client(**kwargs).search(
            _input_ndjson(
                lines,
                size=size,
                sampling_rate=sampling_rate,
                field_resolver=field_resolver,
            ),
            on_done,
            on_error,
            on_always,
            data_type=DataInputType.AUTO,
            **kwargs,
        ):
            yield r

    async def search_csv(
        self,
        lines: Union[Iterable[str], TextIO],
        field_resolver: Optional[Dict[str, str]] = None,
        size: Optional[int] = None,
        sampling_rate: Optional[float] = None,
        on_done: CallbackFnType = None,
        on_error: CallbackFnType = None,
        on_always: CallbackFnType = None,
        **kwargs,
    ):
        """Use a list of lines as the index source for indexing on the current Flow
        :param lines: a list of strings, each is considered as d document
        :param size: the maximum number of the documents
        :param sampling_rate: the sampling rate between [0, 1]
        :param field_resolver: a map from field names defined in ``document`` (JSON, dict) to the field
            names defined in Protobuf. This is only used when the given ``document`` is
            a JSON string or a Python dict.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :yields: results
        """
        from ...clients.sugary_io import _input_csv

        async for r in self._get_client(**kwargs).search(
            _input_csv(
                lines,
                size=size,
                sampling_rate=sampling_rate,
                field_resolver=field_resolver,
            ),
            on_done,
            on_error,
            on_always,
            data_type=DataInputType.AUTO,
            **kwargs,
        ):
            yield r

    @deprecated_alias(
        input_fn=('inputs', 0),
        buffer=('inputs', 1),
        callback=('on_done', 1),
        output_fn=('on_done', 1),
    )
    async def search_lines(
        self,
        lines: Optional[Union[Iterable[str], TextIO]] = None,
        filepath: Optional[str] = None,
        size: Optional[int] = None,
        sampling_rate: Optional[float] = None,
        read_mode: str = 'r',
        line_format: str = 'json',
        field_resolver: Optional[Dict[str, str]] = None,
        on_done: CallbackFnType = None,
        on_error: CallbackFnType = None,
        on_always: CallbackFnType = None,
        **kwargs,
    ):
        """Use a list of files as the query source for searching on the current Flow

        :param filepath: a text file that each line contains a document
        :param lines: a list of strings, each is considered as d document
        :param size: the maximum number of the documents
        :param sampling_rate: the sampling rate between [0, 1]
        :param read_mode: specifies the mode in which the file
                is opened. 'r' for reading in text mode, 'rb' for reading in binary
        :param line_format: the format of each line: ``json`` or ``csv``
        :param field_resolver: a map from field names defined in ``document`` (JSON, dict) to the field
            names defined in Protobuf. This is only used when the given ``document`` is
            a JSON string or a Python dict.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :yields: results
        """
        from ...clients.sugary_io import _input_lines

        async for r in self._get_client(**kwargs).search(
            _input_lines(
                lines,
                filepath,
                size=size,
                sampling_rate=sampling_rate,
                read_mode=read_mode,
                line_format=line_format,
                field_resolver=field_resolver,
            ),
            on_done,
            on_error,
            on_always,
            data_type=DataInputType.CONTENT,
            **kwargs,
        ):
            yield r

    @deprecated_alias(
        input_fn=('inputs', 0),
        buffer=('inputs', 1),
        callback=('on_done', 1),
        output_fn=('on_done', 1),
    )
    async def index(
        self,
        inputs: InputType,
        on_done: CallbackFnType = None,
        on_error: CallbackFnType = None,
        on_always: CallbackFnType = None,
        **kwargs,
    ):
        """Do indexing on the current Flow

        It will start a :py:class:`CLIClient` and call :py:func:`index`.

        :param inputs: An iterator of bytes. If not given, then you have to specify it in **kwargs**.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :yields: results
        """
        async for r in self._get_client(**kwargs).index(
            inputs, on_done, on_error, on_always, **kwargs
        ):
            yield r

    @deprecated_alias(
        input_fn=('inputs', 0),
        buffer=('inputs', 1),
        callback=('on_done', 1),
        output_fn=('on_done', 1),
    )
    async def update(
        self,
        inputs: InputType,
        on_done: CallbackFnType = None,
        on_error: CallbackFnType = None,
        on_always: CallbackFnType = None,
        **kwargs,
    ):
        """Do updates on the current Flow

        It will start a :py:class:`CLIClient` and call :py:func:`index`.

        :param inputs: An iterator of bytes. If not given, then you have to specify it in **kwargs**.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :yields: results
        """
        async for r in self._get_client(**kwargs).update(
            inputs, on_done, on_error, on_always, **kwargs
        ):
            yield r

    @deprecated_alias(
        input_fn=('inputs', 0),
        buffer=('inputs', 1),
        callback=('on_done', 1),
        output_fn=('on_done', 1),
    )
    async def delete(
        self,
        ids: Iterable[str],
        on_done: CallbackFnType = None,
        on_error: CallbackFnType = None,
        on_always: CallbackFnType = None,
        **kwargs,
    ):
        """Do deletion on the current Flow

        :param ids: An iterable of ids
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :yields: results
        """
        async for r in self._get_client(**kwargs).delete(
            ids, on_done, on_error, on_always, **kwargs
        ):
            yield r

    @deprecated_alias(
        input_fn=('inputs', 0),
        buffer=('inputs', 1),
        callback=('on_done', 1),
        output_fn=('on_done', 1),
    )
    async def search(
        self,
        inputs: InputType,
        on_done: CallbackFnType = None,
        on_error: CallbackFnType = None,
        on_always: CallbackFnType = None,
        **kwargs,
    ):
        """Do searching on the current Flow

        It will start a :py:class:`CLIClient` and call :py:func:`search`.

        :param inputs: An iterator of bytes. If not given, then you have to specify it in **kwargs**.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is  is either resolved or rejected.
        :param kwargs: accepts all keyword arguments of `jina client` CLI
        :yields: results
        """
        async for r in self._get_client(**kwargs).search(
            inputs, on_done, on_error, on_always, **kwargs
        ):
            yield r



"""Module for helper functions for clients."""
from typing import Tuple, Sequence

from ... import Document, Request
from ...enums import DataInputType, RequestType
from ...excepts import BadDocType, BadRequestType
from ...excepts import RequestTypeError


def _new_doc_from_data(
    data, data_type: DataInputType, **kwargs
) -> Tuple['Document', 'DataInputType']:
    def _build_doc_from_content():
        with Document(**kwargs) as d:
            d.content = data
        return d, DataInputType.CONTENT

    if data_type == DataInputType.AUTO or data_type == DataInputType.DOCUMENT:
        if isinstance(data, Document):
            # if incoming is already primitive type Document, then all good, best practice!
            return data, DataInputType.DOCUMENT
        try:
            d = Document(data, **kwargs)
            return d, DataInputType.DOCUMENT
        except BadDocType:
            # AUTO has a fallback, now reconsider it as content
            if data_type == DataInputType.AUTO:
                return _build_doc_from_content()
            else:
                raise
    elif data_type == DataInputType.CONTENT:
        return _build_doc_from_content()


def _new_request_from_batch(_kwargs, batch, data_type, mode, queryset, **kwargs):
    req = Request()
    req.request_type = str(mode)

    try:
        # add type-specific fields
        if (
            mode == RequestType.INDEX
            or mode == RequestType.SEARCH
            or mode == RequestType.TRAIN
            or mode == RequestType.UPDATE
        ):
            if 'extra_kwargs' in _kwargs:
                _kwargs.pop('extra_kwargs')  #: data request do not need extra kwargs
            _add_docs_groundtruths(req, batch, data_type, _kwargs)
        elif mode == RequestType.DELETE:
            _add_ids(req, batch)
        elif mode == RequestType.CONTROL:
            _add_control_propagate(req, _kwargs)
        else:
            raise RequestTypeError(
                f'generating request from {mode} is not yet supported'
            )
    except Exception as ex:
        raise BadRequestType(
            f'error when building {req.request_type} from {batch}'
        ) from ex

    # add common fields
    if isinstance(queryset, Sequence):
        req.queryset.extend(queryset)
    elif queryset is not None:
        req.queryset.append(queryset)
    return req


def _add_docs_groundtruths(req, batch, data_type, _kwargs):
    for content in batch:
        if isinstance(content, tuple) and len(content) == 2:
            # content comes in pair,  will take the first as the input and the second as the ground truth

            # note how data_type is cached
            d, data_type = _new_doc_from_data(content[0], data_type, **_kwargs)
            gt, _ = _new_doc_from_data(content[1], data_type, **_kwargs)
            req.docs.append(d)
            req.groundtruths.append(gt)
        else:
            d, data_type = _new_doc_from_data(content, data_type, **_kwargs)
            req.docs.append(d)


def _add_ids(req, batch):
    string_ids = (str(doc_id) for doc_id in batch)
    req.ids.extend(string_ids)


def _add_control_propagate(req, kwargs):
    from ...proto import jina_pb2

    extra_kwargs = kwargs[
        'extra_kwargs'
    ]  #: control command and args are stored inside extra_kwargs
    _available_commands = dict(
        jina_pb2.RequestProto.ControlRequestProto.DESCRIPTOR.enum_values_by_name
    )

    if 'command' in extra_kwargs:
        command = extra_kwargs['command']
    else:
        raise BadRequestType(
            'sending ControlRequest from Client must contain the field `command`'
        )

    if command in _available_commands:
        req.control.command = getattr(
            jina_pb2.RequestProto.ControlRequestProto, command
        )
    else:
        raise ValueError(
            f'command "{command}" is not supported, must be one of {_available_commands}'
        )
    req.targets.extend(extra_kwargs.get('targets', []))
    req.control.propagate = True



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import numpy as np

from jina.executors.decorators import batching, as_ndarray
from jina.executors.encoders.frameworks import BaseTFEncoder

UNIVERSAL_SENTENCE_ENCODER = 'https://tfhub.dev/google/universal-sentence-encoder/4'
MODEL_ENCODER_CMLM = "https://tfhub.dev/google/universal-sentence-encoder-cmlm/en-base/1"
PREPROCESOR_CMLM = "https://tfhub.dev/tensorflow/bert_en_uncased_preprocess/2"


class UniversalSentenceEncoder(BaseTFEncoder):
    """
    Encode an 1d array of string in size `B` into an ndarray in size `B x D`

    The ndarray potentially is BatchSize x (Channel x Height x Width)

    :class:`UniversalSentenceEncoder` is a encoder based on the Universal Sentence
    Encoder family (https://tfhub.dev/google/collections/universal-sentence-encoder/1).

    :param model_url: the url of the model (TensorFlow Hub).
        For supported models see family overview:
        https://tfhub.dev/google/collections/universal-sentence-encoder/1)
    :param args:  Additional positional arguments
    :param kwargs: Additional keyword arguments
    """

    class GeneralEncoder:
        """
        General universal sentence encoder.

        Loads a model and it encodes from an 1d array of string
        in size `B` into an ndarray in size `B x D`.

        :param model_url: the url of the model (TensorFlow Hub).
            For supported models see family overview:
            https://tfhub.dev/google/collections/universal-sentence-encoder/1)
        """

        def __init__(self, model_url: str):
            """Set GeneralEncoder Constructor."""
            import tensorflow_hub as hub
            self.model_url = model_url
            self.model = hub.load(self.model_url)

        def encode(self, content: 'np.ndarray') -> 'np.ndarray':
            """
            Encode ``Document`` content into an ndarray

            :param content: a 1d array of string type in size `B`
            :param args:  Additional positional arguments
            :param kwargs: Additional keyword arguments
            :return: an ndarray in size `B x D`
            """
            return self.model(content)

    class CMLMEncoder:
        """
        Private class encoder to represent a CMLM universal Sentence Encoder family.

        It encodes ``Document`` content from an 1d array of string in size `B`
        into an ndarray in size `B x D`.
        (https://tfhub.dev/google/universal-sentence-encoder-cmlm/en-base/1).
        """

        def __init__(self):
            """Set CMLMEncoder Constructor."""
            import tensorflow_text as text
            import tensorflow_hub as hub
            self.bert_preprocessor = hub.KerasLayer(PREPROCESOR_CMLM)
            self.encoder = hub.KerasLayer(MODEL_ENCODER_CMLM)

        def encode(self, content: 'np.ndarray') -> 'np.ndarray':
            """
            Encode ``Document`` content into an ndarray

            :param content: a 1d array of string type in size `B`
            :param args:  Additional positional arguments
            :param kwargs: Additional keyword arguments
            :return: an ndarray in size `B x D`
            """
            return self.encoder(self.bert_preprocessor(content))['default']

    def __init__(
            self,
            model_url: str = UNIVERSAL_SENTENCE_ENCODER,
            *args,
            **kwargs):
        """Set UniversalSentenceEncoder constructor."""
        super().__init__(*args, **kwargs)
        self.model_url = model_url

    def post_init(self):
        """Load Sentence encoder model"""
        self.to_device()

        if self.model_url == MODEL_ENCODER_CMLM:
            self.sentence_encoder = UniversalSentenceEncoder.CMLMEncoder()
        else:
            self.sentence_encoder = UniversalSentenceEncoder.GeneralEncoder(self.model_url)

    @batching
    @as_ndarray
    def encode(self, content: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        """
        Encode an array of string in size `B` into an ndarray in size `B x D`

        The ndarray potentially is BatchSize x (Channel x Height x Width)

        :param content: a 1d array of string type in size `B`
        :param args:  Additional positional arguments
        :param kwargs: Additional keyword arguments
        :return: an ndarray in size `B x D`
        """
        return self.sentence_encoder.encode(content)
